# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=missing-docstring
# pylint: disable=wrong-import-position
# pylint: disable=no-self-use
# pylint: disable=protected-access
# pylint: disable=function-redefined
# pylint: disable=invalid-name

import logging
import os
import os.path as p
import time
from contextlib import contextmanager
from tempfile import mkdtemp
from threading import Event, Timer
from typing import Any

import parameterized  # type: ignore
import six
import unittest2  # type: ignore
from mock import MagicMock, call, patch
from pyls import _utils  # type: ignore
from pyls import lsp as defines
from pyls import uris
from pyls.workspace import Workspace  # type: ignore
from pyls_jsonrpc.streams import JsonRpcStreamReader  # type: ignore
from tabulate import tabulate

from nose2.tools import such  # type: ignore

from hdl_checker.exceptions import HdlCheckerNotInitializedError
from hdl_checker.parsers.elements.dependency_spec import DependencySpec
from hdl_checker.parsers.elements.design_unit import (
    DesignUnitType,
    VerilogDesignUnit,
    VhdlDesignUnit,
)
from hdl_checker.parsers.elements.identifier import Identifier
from hdl_checker.path import Path
from hdl_checker.types import Location

from hdl_checker.tests import (  # isort:skip
    assertCountEqual,
    getTestTempPath,
    setupTestSuport,
    TestCase,
)


# Debouncing will hurt testing since it won't actually call the debounced
# function if we call it too quickly.
def _debounce(interval_s, keyed_by=None):  # pylint: disable=unused-argument
    def wrapper(func):
        def debounced(*args, **kwargs):
            result = func(*args, **kwargs)
            _logger.info("%s(%s, %s) returned %s", func.__name__, args, kwargs, result)
            return result

        return debounced

    return wrapper


# Mock debounce before it's applied
_utils.debounce = _debounce

from hdl_checker import lsp  # isort:skip
from hdl_checker.diagnostics import CheckerDiagnostic, DiagType  # isort:skip
from hdl_checker.utils import ON_WINDOWS  # isort:skip

_logger = logging.getLogger(__name__)

JSONRPC_VERSION = "2.0"
LSP_MSG_TEMPLATE = {"jsonrpc": JSONRPC_VERSION, "id": 1, "processId": None}

LSP_MSG_EMPTY_RESPONSE = {"jsonrpc": JSONRPC_VERSION, "id": 1, "result": None}

MOCK_WAIT_TIMEOUT = 5

TEST_TEMP_PATH = getTestTempPath(__name__)
TEST_PROJECT = p.join(TEST_TEMP_PATH, "test_project")

if ON_WINDOWS:
    TEST_PROJECT = TEST_PROJECT.lower()


class MockWaitTimeout(Exception):
    def __init__(self, msg):
        super(MockWaitTimeout, self).__init__(self)
        self.msg = msg

    def __str__(self):
        return self.msg


class TestCheckerDiagToLspDict(unittest2.TestCase):
    @parameterized.parameterized.expand(
        [
            (DiagType.INFO, defines.DiagnosticSeverity.Hint),
            (DiagType.STYLE_INFO, defines.DiagnosticSeverity.Hint),
            (DiagType.STYLE_WARNING, defines.DiagnosticSeverity.Information),
            (DiagType.STYLE_ERROR, defines.DiagnosticSeverity.Information),
            (DiagType.WARNING, defines.DiagnosticSeverity.Warning),
            (DiagType.ERROR, defines.DiagnosticSeverity.Error),
            (DiagType.NONE, defines.DiagnosticSeverity.Error),
        ]
    )
    def test_converting_to_lsp(self, diag_type, severity):
        # type: (...) -> Any
        _logger.info("Running %s and %s", diag_type, severity)

        diag = CheckerDiagnostic(
            checker="hdl_checker test",
            text="some diag",
            filename=Path("filename"),
            line_number=1,
            column_number=1,
            error_code="error code",
            severity=diag_type,
        )

        self.assertEqual(
            lsp.checkerDiagToLspDict(diag),
            {
                "source": "hdl_checker test",
                "range": {
                    "start": {"line": 0, "character": 0},
                    "end": {"line": 0, "character": 0},
                },
                "message": "some diag",
                "severity": severity,
                "code": "error code",
            },
        )

    def test_workspace_notify(self):
        # type: (...) -> Any
        workspace = MagicMock(spec=Workspace)

        server = lsp.HdlCodeCheckerServer(
            workspace, root_path=mkdtemp(prefix="hdl_checker_")
        )

        server._handleUiInfo("some info")  # pylint: disable=protected-access
        workspace.show_message.assert_called_once_with(
            "some info", defines.MessageType.Info
        )
        workspace.show_message.reset_mock()

        server._handleUiWarning("some warning")  # pylint: disable=protected-access
        workspace.show_message.assert_called_once_with(
            "some warning", defines.MessageType.Warning
        )
        workspace.show_message.reset_mock()

        server._handleUiError("some error")  # pylint: disable=protected-access
        workspace.show_message.assert_called_once_with(
            "some error", defines.MessageType.Error
        )


such.unittest.TestCase.maxDiff = None

with such.A("LSP server") as it:
    if six.PY2:
        it.assertCountEqual = assertCountEqual(it)

    def _initializeServer(server, params=None):
        it.assertEqual(
            server.m_initialize(**(params or {})),
            {
                "capabilities": {
                    "textDocumentSync": 1,
                    "definitionProvider": True,
                    "hoverProvider": True,
                }
            },
        )

        it.assertEqual(server.m_initialized(), None)

    def startLspServer():
        _logger.debug("Creating server")
        tx_r, tx_w = os.pipe()
        it.tx = JsonRpcStreamReader(os.fdopen(tx_r, "rb"))

        it.rx = MagicMock()
        it.rx.closed = False

        it.server = lsp.HdlCheckerLanguageServer(it.rx, os.fdopen(tx_w, "wb"))

    def stopLspServer():
        _logger.debug("Shutting down server")
        msg = LSP_MSG_TEMPLATE.copy()
        msg.update({"method": "exit"})
        it.server._endpoint.consume(msg)

        # Close the pipe from the server to stdout and empty any pending
        # messages
        it.tx.close()
        it.tx.listen(_logger.fatal)

        del it.server

    def _waitOnMockCall(meth):
        event = Event()

        timer = Timer(MOCK_WAIT_TIMEOUT, event.set)
        timer.start()

        result = []
        while not event.isSet():
            if meth.mock_calls:
                result = meth.mock_calls[0]
                timer.cancel()
                break

            time.sleep(0.1)

        if event.isSet():
            _logger.error("Timed out waiting for %s", meth)
            raise MockWaitTimeout("Timeout waiting for %s" % meth)

        return result

    def checkLintFileOnOpen(source):
        return checkLintFileOnMethod(source, "m_text_document__did_open")

    def checkLintFileOnSave(source):
        return checkLintFileOnMethod(source, "m_text_document__did_save")

    def checkLintFileOnMethod(source, method):
        with patch.object(it.server.workspace, "publish_diagnostics"):
            _logger.info("Sending %s request", method)
            getattr(it.server, method)(
                textDocument={"uri": uris.from_fs_path(source), "text": None}
            )

            mock_call = _waitOnMockCall(it.server.workspace.publish_diagnostics)
            doc_uri, diagnostics = mock_call[1]
            _logger.info("doc_uri: %s", doc_uri)
            _logger.info("diagnostics: %s", diagnostics)

            it.assertEqual(doc_uri, uris.from_fs_path(source))
            return diagnostics

    @it.has_setup
    def setup():
        setupTestSuport(TEST_TEMP_PATH)

    @it.should("not warn when setup takes less than lsp._SETUP_IS_TOO_LONG_TIMEOUT")
    @patch("hdl_checker.lsp.HdlCheckerLanguageServer.showWarning")
    def test(show_warning):
        startLspServer()

        _initializeServer(
            it.server, params={"rootUri": uris.from_fs_path(TEST_PROJECT)}
        )
        show_warning.assert_not_called()
        stopLspServer()

    import hdl_checker

    @it.should(  # type: ignore
        "warn when setup takes more than lsp._SETUP_IS_TOO_LONG_TIMEOUT"
    )
    @patch("hdl_checker.lsp.HdlCheckerLanguageServer.showWarning")
    @patch("hdl_checker.lsp._SETUP_IS_TOO_LONG_TIMEOUT", 0.5)
    @patch.object(
        hdl_checker.base_server.HdlCodeCheckerBase,
        "configure",
        lambda self, _: time.sleep(1),
    )
    def test(show_warning):
        startLspServer()

        _initializeServer(
            it.server, params={"rootUri": uris.from_fs_path(TEST_PROJECT)}
        )
        show_warning.assert_called_once_with(lsp._SETUP_IS_TOO_LONG_MSG)
        stopLspServer()

    @it.should(  # type: ignore
        "show info and warning messages"
    )
    @patch("pyls.workspace.Workspace.show_message")
    def test(show_message):
        startLspServer()

        _initializeServer(
            it.server, params={"rootUri": uris.from_fs_path(TEST_PROJECT)}
        )

        # Initialization calls
        show_message.assert_has_calls(
            [
                call(
                    "Searching %s for HDL files..." % TEST_PROJECT,
                    defines.MessageType.Info,
                ),
                call("Added 10 sources", defines.MessageType.Info),
            ]
        )

        show_message.reset_mock()

        it.server.showWarning("some warning")
        show_message.assert_called_once_with(
            "some warning", defines.MessageType.Warning
        )

        stopLspServer()

    with it.having("root URI set but no project file"):

        @it.has_setup
        def setup():
            startLspServer()

        @it.has_teardown
        def teardown():
            stopLspServer()

        import hdl_checker

        @it.should("search for files on initialization")  # type: ignore
        @patch.object(hdl_checker.base_server.HdlCodeCheckerBase, "configure")
        @patch.object(
            hdl_checker.config_generators.base_generator.BaseGenerator, "generate"
        )
        def test(configure, generate):
            _initializeServer(
                it.server, params={"rootUri": uris.from_fs_path(TEST_PROJECT)}
            )
            configure.assert_called_once()
            generate.assert_called_once()

        @it.should("lint file when opening it")  # type: ignore
        def test():
            source = p.join(TEST_PROJECT, "another_library", "foo.vhd")

            it.assertCountEqual(
                checkLintFileOnOpen(source),
                [
                    {
                        "source": "HDL Checker/static",
                        "range": {
                            "start": {"line": 28, "character": 11},
                            "end": {"line": 28, "character": 11},
                        },
                        "message": "Signal 'neat_signal' is never used",
                        "severity": defines.DiagnosticSeverity.Information,
                    }
                ],
            )

        @it.should("fail to lint a file outside workspace")  # type: ignore
        def test():
            source = p.join(TEST_PROJECT, "basic_library", "clock_divider.vhd")
            with it.assertRaises(MockWaitTimeout):
                checkLintFileOnSave(source)

    with it.having("an existing and valid old style project file"):

        @it.has_setup
        def setup():
            startLspServer()

        @it.has_teardown
        def teardown():
            stopLspServer()

        @it.should("respond capabilities upon initialization")  # type: ignore
        def test():
            _initializeServer(
                it.server,
                params={
                    "rootUri": uris.from_fs_path(TEST_PROJECT),
                    "initializationOptions": {"project_file": "vimhdl.prj"},
                },
            )

        @it.should("lint file when opening it")  # type: ignore
        def test():
            source = p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd")

            it.assertCountEqual(checkLintFileOnOpen(source), [])

    with it.having("a non existing project file"):

        @it.has_setup
        def setup():
            startLspServer()

        @it.has_teardown
        def teardown():
            stopLspServer()

        @it.should("respond capabilities upon initialization")  # type: ignore
        def test():
            it.project_file = "__some_project_file.prj"
            it.assertFalse(p.exists(it.project_file))

            _initializeServer(
                it.server,
                params={
                    "rootUri": uris.from_fs_path(TEST_PROJECT),
                    "initializationOptions": {"project_file": it.project_file},
                },
            )

        @it.should("lint file when opening it")  # type: ignore
        def test():
            source = p.join(TEST_PROJECT, "another_library", "foo.vhd")
            diagnostics = checkLintFileOnOpen(source)

            it.assertCountEqual(
                diagnostics,
                [
                    {
                        "source": "HDL Checker/static",
                        "range": {
                            "start": {"line": 28, "character": 11},
                            "end": {"line": 28, "character": 11},
                        },
                        "message": "Signal 'neat_signal' is never used",
                        "severity": defines.DiagnosticSeverity.Information,
                    }
                ],
            )

    # as per LSP spec, this is invalid. rootUri is a mandatory field
    with it.having("neither root URI nor project file set"):

        @it.has_setup
        def setup():
            startLspServer()

        @it.has_teardown
        def teardown():
            stopLspServer()

        @it.should("respond capabilities upon initialization")  # type: ignore
        def test():
            _initializeServer(
                it.server, params={"initializationOptions": {"project_file": None}}
            )

        # Linting might happen on a different thread, so we might not actually
        # get the exception. In any case, its constructor must be called
        @it.should("fail to lint files")  # type: ignore
        @patch.object(HdlCheckerNotInitializedError, "__init__", return_value=None)
        def test(exc):
            source = p.join(TEST_PROJECT, "basic_library", "clock_divider.vhd")
            with it.assertRaises(
                HdlCheckerNotInitializedError if ON_WINDOWS else MockWaitTimeout
            ):
                checkLintFileOnOpen(source)

            exc.assert_called_once()

    with it.having("no root URI but project file set"):

        @it.has_setup
        def setup():
            startLspServer()

        @it.has_teardown
        def teardown():
            stopLspServer()

        @it.should("respond capabilities upon initialization")  # type: ignore
        def test():
            # In this case, project file is an absolute path, since there's no
            # root URI
            _initializeServer(
                it.server,
                params={
                    "initializationOptions": {
                        "project_file": p.join(TEST_PROJECT, "vimhdl.prj")
                    }
                },
            )

        # Linting might happen on a different thread, so we might not actually
        # get the exception. In any case, its constructor must be called
        @it.should("fail to lint files")  # type: ignore
        @patch.object(HdlCheckerNotInitializedError, "__init__", return_value=None)
        def test(exc):
            source = p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd")
            with it.assertRaises(
                HdlCheckerNotInitializedError if ON_WINDOWS else MockWaitTimeout
            ):
                checkLintFileOnOpen(source)

            exc.assert_called_once()


class TestValidProject(TestCase):

    params = {
        "rootUri": uris.from_fs_path(TEST_PROJECT),
        "initializationOptions": {"project_file": "config.json"},
    }

    def setUp(self):
        setupTestSuport(TEST_TEMP_PATH)

        _logger.debug("Creating server")
        tx_r, tx_w = os.pipe()
        self.tx_stream_reader = JsonRpcStreamReader(os.fdopen(tx_r, "rb"))

        rx_stream = MagicMock()
        rx_stream.closed = False

        self.server = lsp.HdlCheckerLanguageServer(rx_stream, os.fdopen(tx_w, "wb"))

        # Initialize server
        _logger.info("Calling m_initialize")
        self.assertEqual(
            self.server.m_initialize(**(self.params or {})),
            {
                "capabilities": {
                    "textDocumentSync": 1,
                    "definitionProvider": True,
                    "hoverProvider": True,
                }
            },
        )

        _logger.info("Calling m_initialized")
        self.assertIsNone(self.server.m_initialized())

    def teardown(self):
        _logger.debug("Shutting down server")
        msg = LSP_MSG_TEMPLATE.copy()
        msg.update({"method": "exit"})
        self.server._endpoint.consume(msg)

        # Close the pipe from the server to stdout and empty any pending
        # messages
        self.tx_stream_reader.close()
        self.tx_stream_reader.listen(_logger.fatal)

        del self.server

    def _checkLintFileOnOpen(self, source):
        return self._checkLintFileOnMethod(source, "m_text_document__did_open")

    def _checkLintFileOnSave(self, source):
        return self._checkLintFileOnMethod(source, "m_text_document__did_save")

    def _checkLintFileOnMethod(self, source, method):
        with patch.object(self.server.workspace, "publish_diagnostics"):
            _logger.info("Sending %s request", method)
            getattr(self.server, method)(
                textDocument={"uri": uris.from_fs_path(source), "text": None}
            )

            mock_call = _waitOnMockCall(self.server.workspace.publish_diagnostics)
            doc_uri, diagnostics = mock_call[1]
            _logger.info("doc_uri: %s", doc_uri)
            _logger.info("diagnostics: %s", diagnostics)

            self.assertEqual(doc_uri, uris.from_fs_path(source))
            return diagnostics

    def test_LintFileOnOpening(self):
        source = p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd")
        self.assertEqual(self._checkLintFileOnOpen(source), [])

    def runTestBuildSequenceTable(self, tablefmt):
        very_common_pkg = Path(
            p.join(TEST_PROJECT, "basic_library", "very_common_pkg.vhd")
        )
        clk_en_generator = Path(
            p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd")
        )

        expected = [
            "Build sequence for %s is" % str(clk_en_generator),
            "",
            tabulate(
                [
                    (1, "basic_library", str(very_common_pkg)),
                    (2, "library", str(clk_en_generator)),
                ],
                headers=("#", "Library", "Path"),
                tablefmt=tablefmt,
            ),
        ]

        self.assertEqual(
            self.server._getBuildSequenceForHover(clk_en_generator), "\n".join(expected)
        )

    @patch("hdl_checker.lsp.HdlCheckerLanguageServer._use_markdown_for_hover", 0)
    @patch(
        "hdl_checker.builders.base_builder.BaseBuilder.builtin_libraries",
        [Identifier("ieee")],
    )
    def test_ReportBuildSequencePlain(self):
        self.runTestBuildSequenceTable(tablefmt="plain")

    @patch("hdl_checker.lsp.HdlCheckerLanguageServer._use_markdown_for_hover", 1)
    @patch(
        "hdl_checker.builders.base_builder.BaseBuilder.builtin_libraries",
        [Identifier("ieee")],
    )
    def test_ReportBuildSequenceMarkdown(self):
        self.runTestBuildSequenceTable(tablefmt="github")

    @patch.object(
        hdl_checker.base_server.HdlCodeCheckerBase,
        "resolveDependencyToPath",
        lambda self, _: None,
    )
    def test_DependencyInfoForPathNotFound(self):
        path = Path(p.join(TEST_PROJECT, "another_library", "foo.vhd"))
        dependency = DependencySpec(
            name=Identifier("clock_divider"),
            library=Identifier("basic_library"),
            owner=path,
            locations=(),
        )
        self.assertEqual(
            self.server._getDependencyInfoForHover(dependency),
            "Couldn't find a source defining 'basic_library.clock_divider'",
        )

    @patch.object(
        hdl_checker.base_server.HdlCodeCheckerBase,
        "resolveDependencyToPath",
        lambda self, _: (Path("some_path"), Identifier("some_library")),
    )
    def test_ReportDependencyInfo(self):
        path = Path(p.join(TEST_PROJECT, "another_library", "foo.vhd"))
        dependency = DependencySpec(
            name=Identifier("clock_divider"),
            library=Identifier("basic_library"),
            owner=path,
            locations=(),
        )
        self.assertEqual(
            self.server._getDependencyInfoForHover(dependency),
            'Path "some_path", library "some_library"',
        )

    def test_ReportDesignUnitAccordingToPosition(self):
        UNIT_A = VhdlDesignUnit(
            owner=Path(p.join(TEST_PROJECT, "another_library", "foo.vhd")),
            type_=DesignUnitType.entity,
            name="unit_a",
            locations=(Location(line=1, column=2), Location(line=3, column=4)),
        )

        UNIT_B = VerilogDesignUnit(
            owner=Path(p.join(TEST_PROJECT, "another_library", "foo.vhd")),
            type_=DesignUnitType.package,
            name="unit_b",
            locations=(Location(line=5, column=6), Location(line=7, column=8)),
        )

        DEP_A = DependencySpec(
            name=Identifier("dep_a"),
            library=Identifier("lib_a"),
            owner=Path(p.join(TEST_PROJECT, "another_library", "foo.vhd")),
            locations=(Location(line=9, column=10), Location(line=11, column=12)),
        )

        DEP_B = DependencySpec(
            name=Identifier("dep_a"),
            library=Identifier("lib_a"),
            owner=Path(p.join(TEST_PROJECT, "another_library", "foo.vhd")),
            locations=(Location(line=13, column=14), Location(line=15, column=16)),
        )

        def getDesignUnitsByPath(self, path):  # pylint: disable=unused-argument
            if path != Path(p.join(TEST_PROJECT, "another_library", "foo.vhd")):
                it.fail("Expected foo.vhd but got %s" % path)
            return {UNIT_A, UNIT_B}

        def getDependenciesByPath(self, path):  # pylint: disable=unused-argument
            if path != Path(p.join(TEST_PROJECT, "another_library", "foo.vhd")):
                it.fail("Expected foo.vhd but got %s" % path)
            return {DEP_A, DEP_B}

        @contextmanager
        def patchServer():
            with patch.object(
                hdl_checker.database.Database,
                "getDesignUnitsByPath",
                getDesignUnitsByPath,
            ):
                with patch.object(
                    hdl_checker.database.Database,
                    "getDependenciesByPath",
                    getDependenciesByPath,
                ):
                    yield

        path = Path(p.join(TEST_PROJECT, "another_library", "foo.vhd"))

        with patchServer():
            # Check locations outside return nothing
            self.assertIsNone(self.server._getElementAtPosition(path, Location(0, 0)))

            # Check design units are found, ensure boundaries match
            self.assertIsNone(self.server._getElementAtPosition(path, Location(1, 1)))
            self.assertIs(
                self.server._getElementAtPosition(path, Location(1, 2)), UNIT_A
            )
            self.assertIs(
                self.server._getElementAtPosition(path, Location(1, 7)), UNIT_A
            )
            self.assertIsNone(self.server._getElementAtPosition(path, Location(1, 8)))

            self.assertIsNone(self.server._getElementAtPosition(path, Location(3, 3)))
            self.assertIs(
                self.server._getElementAtPosition(path, Location(3, 4)), UNIT_A
            )
            self.assertIs(
                self.server._getElementAtPosition(path, Location(3, 9)), UNIT_A
            )
            self.assertIsNone(self.server._getElementAtPosition(path, Location(3, 10)))

            self.assertIsNone(self.server._getElementAtPosition(path, Location(5, 5)))
            self.assertIs(
                self.server._getElementAtPosition(path, Location(5, 6)), UNIT_B
            )
            self.assertIs(
                self.server._getElementAtPosition(path, Location(5, 11)), UNIT_B
            )
            self.assertIsNone(self.server._getElementAtPosition(path, Location(5, 12)))

            self.assertIsNone(self.server._getElementAtPosition(path, Location(7, 7)))
            self.assertIs(
                self.server._getElementAtPosition(path, Location(7, 8)), UNIT_B
            )
            self.assertIs(
                self.server._getElementAtPosition(path, Location(7, 13)), UNIT_B
            )
            self.assertIsNone(self.server._getElementAtPosition(path, Location(7, 14)))

            # Now check dependencies
            self.assertIsNone(self.server._getElementAtPosition(path, Location(9, 9)))
            self.assertIs(
                self.server._getElementAtPosition(path, Location(9, 10)), DEP_A
            )
            self.assertIs(
                self.server._getElementAtPosition(path, Location(9, 20)), DEP_A
            )
            self.assertIsNone(self.server._getElementAtPosition(path, Location(9, 21)))

            self.assertIsNone(self.server._getElementAtPosition(path, Location(11, 11)))
            self.assertIs(
                self.server._getElementAtPosition(path, Location(11, 12)), DEP_A
            )
            self.assertIs(
                self.server._getElementAtPosition(path, Location(11, 22)), DEP_A
            )
            self.assertIsNone(self.server._getElementAtPosition(path, Location(11, 23)))

            self.assertIsNone(self.server._getElementAtPosition(path, Location(13, 13)))
            self.assertIs(
                self.server._getElementAtPosition(path, Location(13, 14)), DEP_B
            )
            self.assertIs(
                self.server._getElementAtPosition(path, Location(13, 24)), DEP_B
            )
            self.assertIsNone(self.server._getElementAtPosition(path, Location(13, 25)))

            self.assertIsNone(self.server._getElementAtPosition(path, Location(15, 15)))
            self.assertIs(
                self.server._getElementAtPosition(path, Location(15, 16)), DEP_B
            )
            self.assertIs(
                self.server._getElementAtPosition(path, Location(15, 26)), DEP_B
            )
            self.assertIsNone(self.server._getElementAtPosition(path, Location(15, 27)))

    @patch(
        "hdl_checker.builders.base_builder.BaseBuilder.builtin_libraries",
        [Identifier("ieee")],
    )
    def test_HoverOnInvalidRange(self):
        path = p.join(TEST_PROJECT, "another_library", "foo.vhd")
        self.assertEqual(
            self.server.hover(uris.from_fs_path(path), {"line": 0, "character": 0}), {}
        )

    @patch(
        "hdl_checker.builders.base_builder.BaseBuilder.builtin_libraries",
        [Identifier("ieee")],
    )
    def test_HoverOnDesignUnit(self):
        path_to_foo = p.join(TEST_PROJECT, "another_library", "foo.vhd")
        very_common_pkg = p.join(TEST_PROJECT, "basic_library", "very_common_pkg.vhd")
        package_with_constants = p.join(
            TEST_PROJECT, "basic_library", "package_with_constants.vhd"
        )
        clock_divider = p.join(TEST_PROJECT, "basic_library", "clock_divider.vhd")

        expected = [
            "Build sequence for %s is" % str(path_to_foo),
            "",
            tabulate(
                [
                    (1, "basic_library", str(very_common_pkg)),
                    (2, "basic_library", str(package_with_constants)),
                    (3, "basic_library", str(clock_divider)),
                    (4, "library", str(path_to_foo)),
                ],
                headers=("#", "Library", "Path"),
                tablefmt="plain",
            ),
        ]

        self.assertDictEqual(
            self.server.hover(
                uris.from_fs_path(path_to_foo), {"line": 7, "character": 7}
            ),
            {"contents": "\n".join(expected)},
        )

    @patch(
        "hdl_checker.builders.base_builder.BaseBuilder.builtin_libraries",
        [Identifier("ieee")],
    )
    def test_HoverOnDependency(self):
        path_to_foo = p.join(TEST_PROJECT, "another_library", "foo.vhd")
        clock_divider = p.join(TEST_PROJECT, "basic_library", "clock_divider.vhd")

        self.assertDictEqual(
            self.server.hover(
                uris.from_fs_path(path_to_foo), {"line": 32, "character": 32}
            ),
            {"contents": 'Path "%s", library "basic_library"' % clock_divider},
        )

    @patch(
        "hdl_checker.builders.base_builder.BaseBuilder.builtin_libraries",
        [Identifier("ieee")],
    )
    def test_GetDefinitionMatchingDependency(self):
        source = p.join(TEST_PROJECT, "basic_library", "use_entity_a_and_b.vhd")
        target = p.join(TEST_PROJECT, "basic_library", "two_entities_one_file.vhd")

        definitions = self.server.definitions(
            uris.from_fs_path(source), {"line": 1, "character": 9}
        )

        self.assertIn(
            {
                "uri": uris.from_fs_path(target),
                "range": {
                    "start": {"line": 1, "character": 7},
                    "end": {"line": 1, "character": 15},
                },
            },
            definitions,
        )

        self.assertIn(
            {
                "uri": uris.from_fs_path(target),
                "range": {
                    "start": {"line": 4, "character": 7},
                    "end": {"line": 4, "character": 15},
                },
            },
            definitions,
        )

    @patch(
        "hdl_checker.builders.base_builder.BaseBuilder.builtin_libraries",
        [Identifier("ieee")],
    )
    def test_GetDefinitionBuiltInLibrary(self):
        path_to_foo = p.join(TEST_PROJECT, "another_library", "foo.vhd")

        self.assertEqual(
            self.server.definitions(
                uris.from_fs_path(path_to_foo), {"line": 3, "character": 15}
            ),
            [],
        )

    @patch(
        "hdl_checker.builders.base_builder.BaseBuilder.builtin_libraries",
        [Identifier("ieee")],
    )
    def test_GetDefinitionNotKnown(self):
        path_to_foo = p.join(TEST_PROJECT, "another_library", "foo.vhd")

        self.assertEqual(
            self.server.definitions(
                uris.from_fs_path(path_to_foo), {"line": 0, "character": 0}
            ),
            [],
        )


it.createTests(globals())



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.

import logging
from typing import Iterable, Optional, Set
import abc

from hdl_checker.types import Location
from hdl_checker.utils import HashableByKey

_logger = logging.getLogger(__name__)

LocationList = Iterable[Location]


class ParsedElement(HashableByKey):
    __metaclass__ = abc.ABCMeta

    def __init__(self, locations=None):
        # type: (Optional[LocationList]) -> None
        set_of_locations = set()  # type: Set[Location]
        for line_number, column_number in locations or []:
            set_of_locations.add(
                Location(
                    None if line_number is None else int(line_number),
                    None if column_number is None else int(column_number),
                )
            )

        self._locations = tuple(set_of_locations)

    @property
    def locations(self):
        return self._locations

    @property
    def __hash_key__(self):
        return (self.locations,)

    @abc.abstractmethod
    def __len__(self):
        # type: (...) -> int
        """
        len(self) should return the length the parsed element uses on the text.
        It will be used to calculate an end position for it and allow checking
        if a given location is within the element's text
        """

    def includes(self, position):
        # type: (Location) -> bool
        name_length = len(self)

        for location in self.locations:
            if position.line != location.line:
                continue

            if location.column <= position.column < location.column + name_length:
                return True
        return False



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=function-redefined
# pylint: disable=missing-docstring
# pylint: disable=protected-access
# pylint: disable=invalid-name

import logging
import os
import os.path as p

import six
from webtest import TestApp  # type: ignore

from nose2.tools import such  # type: ignore

from hdl_checker.tests import assertCountEqual, disableVunit, getTestTempPath, setupTestSuport

import hdl_checker
import hdl_checker.handlers as handlers
from hdl_checker.diagnostics import (
    CheckerDiagnostic,
    DiagType,
    ObjectIsNeverUsed,
    StaticCheckerDiag,
)
from hdl_checker.parsers.elements.identifier import Identifier
from hdl_checker.path import Path

try:  # Python 3.x
    import unittest.mock as mock  # pylint: disable=import-error, no-name-in-module
except ImportError:  # Python 2.x
    import mock  # type: ignore


TEST_TEMP_PATH = getTestTempPath(__name__)
TEST_PROJECT = p.abspath(p.join(TEST_TEMP_PATH, "test_project"))

SERVER_LOG_LEVEL = os.environ.get("SERVER_LOG_LEVEL", "INFO")

_logger = logging.getLogger(__name__)
HDL_CHECKER_BASE_PATH = p.abspath(p.join(p.dirname(__file__), "..", ".."))


def _path(*args):
    # type: (str) -> str
    "Helper to reduce foorprint of p.join(TEST_TEMP_PATH, *args)"
    return p.join(TEST_TEMP_PATH, *args)


with such.A("hdl_checker bottle app") as it:
    # Workaround for Python 2.x and 3.x differences
    if six.PY2:
        it.assertCountEqual = assertCountEqual(it)

    @it.has_setup
    def setup():
        setupTestSuport(TEST_TEMP_PATH)

        it.project_file = p.join(TEST_PROJECT, "vimhdl.prj")
        it.app = TestApp(handlers.app)

    @it.should("get diagnose info without any project")
    @disableVunit
    def test():
        reply = it.app.post_json("/get_diagnose_info")
        it.assertCountEqual(
            reply.json["info"],
            [
                u"hdl_checker version: %s" % hdl_checker.__version__,
                u"Server PID: %d" % os.getpid(),
                u"Builder: none",
            ],
        )

    @it.should("get diagnose info with an existing project file")  # type: ignore
    @disableVunit
    def test():
        reply = it.app.post("/get_diagnose_info", {"project_file": it.project_file})

        _logger.info("Reply is %s", reply.json["info"])

        it.assertCountEqual(
            reply.json["info"],
            [
                u"hdl_checker version: %s" % hdl_checker.__version__,
                u"Server PID: %d" % os.getpid(),
                u"Builder: none",
            ],
        )

    @it.should("get diagnose info with a non existing project file")  # type: ignore
    @disableVunit
    def test():
        open(_path("foo_bar.prj"), "w").write("")

        reply = it.app.post(
            "/get_diagnose_info", {"project_file": _path("foo_bar.prj")}
        )

        _logger.info("Reply is %s", reply.json["info"])
        it.assertCountEqual(
            reply.json["info"],
            [
                u"hdl_checker version: %s" % hdl_checker.__version__,
                u"Server PID: %d" % os.getpid(),
                u"Builder: none",
            ],
        )

    @it.should("shutdown the server when requested")  # type: ignore
    @disableVunit
    def test():
        open(_path("some_project"), "w").write("")
        # Ensure the server is active
        reply = it.app.post(
            "/get_diagnose_info", {"project_file": _path("some_project")}
        )
        it.assertEqual(reply.status, "200 OK")

        # Send a request to shutdown the server and check if it
        # calls the terminate process method
        pids = []

        with mock.patch("hdl_checker.handlers.terminateProcess", pids.append):
            reply = it.app.post("/shutdown")

        it.assertEqual(pids, [os.getpid()])

    @it.should("rebuild the project with directory cleanup")  # type: ignore
    @disableVunit
    def test():
        project_file = _path("hello.prj")
        open(_path("hello.prj"), "w").write("")
        server = mock.MagicMock()

        servers = mock.MagicMock()
        servers.__getitem__.side_effect = {
            Path(p.dirname(project_file)): server
        }.__getitem__

        with mock.patch.object(hdl_checker.handlers, "servers", servers):
            it.app.post("/rebuild_project", {"project_file": project_file})

        # Check the object was removed from the servers list
        servers.__delitem__.assert_called_once_with(project_file)
        # Check the original server cleaned things up
        server.clean.assert_called_once()

    @it.should("get messages with content")  # type: ignore
    def test():
        data = {
            "project_file": it.project_file,
            "path": p.join(TEST_PROJECT, "another_library", "foo.vhd"),
            "content": "-- TODO: Nothing to see here",
        }

        ui_reply = it.app.post("/get_ui_messages", data)
        reply = it.app.post("/get_messages_by_path", data)

        _logger.info("UI reply: %s", ui_reply)
        _logger.info("Reply: %s", reply)

        messages = [CheckerDiagnostic.fromDict(x) for x in reply.json["messages"]]

        it.assertIn(data["path"], [x.filename for x in messages])

        expected = StaticCheckerDiag(
            filename=data["path"],
            line_number=1,
            column_number=4,
            text="TODO: Nothing to see here",
            severity=DiagType.STYLE_INFO,
        )

        it.assertIn(expected, messages)

    @it.should("get messages by path")  # type: ignore
    def test():
        filename = p.join(TEST_PROJECT, "basic_library", "clock_divider.vhd")
        data = {"project_file": it.project_file, "path": filename}

        ui_reply = it.app.post("/get_ui_messages", data)
        reply = it.app.post("/get_messages_by_path", data)

        _logger.info("UI reply: %s", ui_reply)
        _logger.info("Reply: %s", reply)

        messages = [CheckerDiagnostic.fromDict(x) for x in reply.json["messages"]]

        it.assertCountEqual(
            messages,
            [
                ObjectIsNeverUsed(
                    filename=filename,
                    line_number=27,
                    column_number=12,
                    object_type="signal",
                    object_name="clk_enable_unused",
                )
            ],
        )

    @it.should("get source dependencies")  # type: ignore
    @disableVunit
    def test():
        data = {
            "project_file": it.project_file,
            "path": p.join(TEST_PROJECT, "another_library", "foo.vhd"),
        }

        for _ in range(10):
            ui_reply = it.app.post("/get_ui_messages", data)
            reply = it.app.post("/get_dependencies", data)

            _logger.info("UI reply: %s", ui_reply)
            _logger.info("Reply: %s", reply)

        dependencies = reply.json["dependencies"]

        _logger.info("Dependencies: %s", ", ".join(dependencies))

        it.assertCountEqual(
            ["ieee.std_logic_1164", "ieee.numeric_std", "basic_library.clock_divider"],
            dependencies,
        )

    @it.should("get source build sequence")  # type: ignore
    def test():
        data = {
            "project_file": it.project_file,
            "path": p.join(TEST_PROJECT, "another_library", "foo.vhd"),
        }

        @property
        def builtin_libraries(_):
            return {Identifier("ieee")}

        such.unittest.TestCase.maxDiff = None
        with mock.patch.object(
            hdl_checker.builders.base_builder.BaseBuilder,
            "builtin_libraries",
            builtin_libraries,
        ):
            reply = it.app.post("/get_build_sequence", data)
            #  reply = it.app.post("/shutdown")
            sequence = reply.json["sequence"]

        _logger.info("Sequence: %s", sequence)

        very_common_pkg = p.join(TEST_PROJECT, "basic_library", "very_common_pkg.vhd")
        package_with_constants = p.join(
            TEST_PROJECT, "basic_library", "package_with_constants.vhd"
        )
        clock_divider = p.join(TEST_PROJECT, "basic_library", "clock_divider.vhd")

        it.assertEqual(
            sequence,
            [
                "%s (library: basic_library)" % very_common_pkg,
                "%s (library: basic_library)" % package_with_constants,
                "%s (library: basic_library)" % clock_divider,
            ],
        )


it.createTests(globals())



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"""
hdl_checker provides a Python API between a VHDL project and some HDL
compilers to catch errors and warnings the compilers generate that can
be used to populate syntax checkers and linters of text editors. It
takes into account the sources dependencies when building so you don't
need to provide a source list ordered by hand.
"""
import os

from ._version import get_versions

from hdl_checker.utils import ON_WINDOWS

__author__ = "Andre Souto (andre820@gmail.com)"
__license__ = "GPLv3"
__status__ = "Development"

__version__ = get_versions()["version"]
del get_versions

DEFAULT_PROJECT_FILE = os.environ.get(
    "HDL_CHECKER_DEFAULT_PROJECT_FILE", "_hdl_checker.config" if ON_WINDOWS else ".hdl_checker.config"
)

CACHE_NAME = os.environ.get("HDL_CHECKER_CACHE_NAME", "cache.json")
WORK_PATH = os.environ.get("HDL_CHECKER_WORK_PATH", "_hdl_checker" if ON_WINDOWS else ".hdl_checker")



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Fallback builder for cases where no builder is found"

from hdl_checker.builders.base_builder import BaseBuilder
from hdl_checker.types import FileType


class Fallback(BaseBuilder):
    "Dummy fallback builder"

    # Implementation of abstract class properties
    builder_name = "fallback"
    file_types = {FileType.vhdl, FileType.verilog, FileType.systemverilog}

    def __init__(self, *args, **kwargs):
        # type: (...) -> None
        self._version = "<undefined>"
        super(Fallback, self).__init__(*args, **kwargs)

    # Since Fallback._buildSource returns nothing,
    # Fallback._makeRecords is never called
    def _makeRecords(self, _):  # pragma: no cover
        return []

    def _shouldIgnoreLine(self, line):  # pragma: no cover
        return True

    def _checkEnvironment(self):
        return

    @staticmethod
    def isAvailable():
        return True

    def _buildSource(self, path, library, flags=None):  # pragma: no cover
        return [], []

    def _createLibrary(self, library):  # pragma: no cover
        pass



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Class defining a parsed design unit"

import logging
from typing import Optional, Union

from .identifier import (  # pylint: disable=unused-import
    Identifier,
    VerilogIdentifier,
    VhdlIdentifier,
)
from .parsed_element import LocationList, ParsedElement  # pylint: disable=unused-import

from hdl_checker.path import Path  # pylint: disable=unused-import
from hdl_checker.types import DesignUnitType  # pylint: disable=unused-import

_logger = logging.getLogger(__name__)


class _DesignUnit(ParsedElement):
    """
    Specifies a design unit (uses mostly VHDL nomenclature)
    """

    def __init__(self, owner, type_, name, locations=None):
        # type: (Path, DesignUnitType, Identifier, Optional[LocationList]) -> None
        self._owner = owner
        self._type = type_
        self._name = name

        super(_DesignUnit, self).__init__(locations)

    def __len__(self):
        return len(self.name)

    def __repr__(self):
        return '{}(name="{}", type={}, owner={}, locations={})'.format(
            self.__class__.__name__,
            repr(self.name),
            repr(self.type_),
            repr(self.owner),
            self.locations,
        )

    def __str__(self):
        return "{}(name='{}', type={}, owner='{}')".format(
            self.__class__.__name__, self.name.display_name, self.type_, str(self.owner)
        )

    def __jsonEncode__(self):
        return {
            "owner": self.owner,
            "type_": self.type_,
            "name": self.name,
            "locations": tuple(self.locations),
        }

    @classmethod
    def __jsonDecode__(cls, state):
        """Returns an object of cls based on a given state"""
        # pylint: disable=protected-access
        return cls(
            state.pop("owner"),
            state.pop("type_"),
            state.pop("name"),
            state.pop("locations", None),
        )

    @property
    def owner(self):
        # type: () -> Path
        "Owner of the design unit"
        return self._owner

    @property
    def type_(self):
        # type: () -> DesignUnitType
        "Design unit type"
        return self._type

    @property
    def name(self):
        # type: () -> Identifier
        "Design unit name"
        return self._name

    @property
    def __hash_key__(self):
        return (
            self.owner,
            self.type_,
            self.name,
            super(_DesignUnit, self).__hash_key__,
        )


class VhdlDesignUnit(_DesignUnit):
    """
    Specifies a design unit whose name is case insensitive
    """

    def __init__(self, owner, type_, name, locations=None):
        # type: (Path, DesignUnitType, str, Optional[LocationList]) -> None
        super(VhdlDesignUnit, self).__init__(
            owner=owner, type_=type_, name=VhdlIdentifier(name), locations=locations
        )


class VerilogDesignUnit(_DesignUnit):
    """
    Specifies a design unit whose name is case sensitive
    """

    def __init__(self, owner, type_, name, locations=None):
        # type: (Path, DesignUnitType, str, Optional[LocationList]) -> None
        super(VerilogDesignUnit, self).__init__(
            owner=owner, type_=type_, name=VerilogIdentifier(name), locations=locations
        )


tAnyDesignUnit = Union[VhdlDesignUnit, VerilogDesignUnit]



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Base class for creating a project file"

import abc
import logging
from typing import Dict, Optional, Set, Tuple

from hdl_checker.builder_utils import AnyValidBuilder
from hdl_checker.path import Path
from hdl_checker.types import BuildFlags, FileType

_SOURCE_EXTENSIONS = "vhdl", "sv", "v"
_HEADER_EXTENSIONS = "vh", "svh"

_DEFAULT_LIBRARY_NAME = {"vhdl": "lib", "verilog": "lib", "systemverilog": "lib"}

SourceSpec = Tuple[Path, BuildFlags, Optional[str]]


class BaseGenerator:
    """
    Base class implementing creation of config file semi automatically
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self):  # type: () -> None
        """
        """
        self._logger = logging.getLogger(self.__class__.__name__)
        self._sources = set()  # type: Set[SourceSpec]
        self._include_paths = {
            FileType.verilog: set(),
            FileType.systemverilog: set(),
        }  # type: Dict[FileType, Set[str]]

    def _addSource(self, path, flags=None, library=None):
        # type: (Path, BuildFlags, Optional[str]) -> None
        """
        Add a source to project. 'flags' and 'library' are only used for
        regular sources and not for header files (files ending in .vh or .svh)
        """
        self._logger.debug(
            "Adding path %s (flags=%s, library=%s)", path, flags, library
        )

        if path.basename.split(".")[-1].lower() in ("vh", "svh"):
            file_type = FileType.fromPath(path)
            if file_type in (FileType.verilog, FileType.systemverilog):
                self._include_paths[file_type].add(path.dirname)
        else:
            self._sources.add((path, flags or (), library))

    @abc.abstractmethod
    def _populate(self):  # type: () -> None
        """
        Method that will be called for generating the project file contets and
        should be implemented by child classes
        """

    def _getPreferredBuilder(self):  # type: () -> AnyValidBuilder
        """
        Method should be overridden by child classes to express the preferred
        builder
        """
        return NotImplemented

    def generate(self):
        """
        Runs the child class algorithm to populate the parent object with the
        project info and writes the result to the project file
        """

        self._populate()

        project = {"sources": []}

        builder = self._getPreferredBuilder()
        if builder is not NotImplemented:
            project["builder"] = builder

        # Add include paths if they exists. Need to iterate sorted keys to
        # generate results always in the same order
        for lang in (FileType.verilog, FileType.systemverilog):
            paths = self._include_paths[lang]
            if paths:
                if lang.value not in project:
                    project[lang.value] = {}
                project[lang.value]["include_paths"] = tuple(paths)

        for path, flags, library in self._sources:
            info = {}
            if library:
                info["library"] = library
            if flags:
                info["flags"] = flags

            if info:
                project["sources"].append(
                    (str(path), {"library": library, "flags": tuple(flags)})
                )
            else:
                project["sources"].append(str(path))

        from pprint import pformat

        self._logger.info("Resulting project:\n%s", pformat(project))

        return project



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=function-redefined
# pylint: disable=missing-docstring
# pylint: disable=protected-access

import json
import logging
import os
import os.path as p
import time

import six

from nose2.tools import such  # type: ignore

from hdl_checker.tests import assertCountEqual, assertSameFile, writeListToFile

from hdl_checker.parsers.elements.dependency_spec import DependencySpec
from hdl_checker.parsers.elements.design_unit import VhdlDesignUnit
from hdl_checker.parsers.elements.identifier import Identifier
from hdl_checker.parsers.vhdl_parser import VhdlParser
from hdl_checker.path import Path
from hdl_checker.serialization import StateEncoder, jsonObjectHook
from hdl_checker.types import DesignUnitType

_logger = logging.getLogger(__name__)

TEST_SUPPORT_PATH = p.join(os.environ["TOX_ENV_DIR"], "tmp")
_FILENAME = p.join(TEST_SUPPORT_PATH, "source.vhd")

such.unittest.TestCase.maxDiff = None


def _DependencySpec(owner, name, library, locations=None):
    if library is not None:
        library = Identifier(library, case_sensitive=False)
    return DependencySpec(
        owner=owner,
        name=Identifier(name, case_sensitive=False),
        library=library,
        locations=locations,
    )


with such.A("VHDL source file object") as it:

    it.assertSameFile = assertSameFile(it)

    if six.PY2:
        # Can't use assertCountEqual for lists of unhashable types.
        # Workaround for https://bugs.python.org/issue10242
        it.assertCountEqual = assertCountEqual(it)

    with it.having("an entity code"):

        @it.has_setup
        def setup():
            if os.path.exists(_FILENAME):
                os.remove(_FILENAME)
            it._code = [
                "library ieee ;",
                "use ieee.std_logic_1164.all;",
                "USE  IEEE.STD_LOGIC_ARITH.ALL;",
                "",
                "library work;",
                " use work.package_with_constants;",
                " use  work.cherry_pick.one;",
                "use work.cherry_pick.two;",
                "",
                "-- library foo;",
                "library lib1,lib2;",
                "library lib3, lib4;",
                "library lib5; library lib6;",
                "",
                "entity clock_divider is",
                "    generic (",
                "        DIVIDER : integer := 10",
                "    );",
                "    port (",
                "        reset : in std_logic;",
                "        clk_input : in  std_logic;",
                "        clk_output : out std_logic",
                "    );",
                "end clock_divider;",
                "",
                "architecture clock_divider of clock_divider is",
                "",
                "begin",
                "",
                "end clock_divider;",
            ]

            writeListToFile(_FILENAME, it._code)

        @it.has_teardown
        def teardown():
            if os.path.exists(_FILENAME):
                os.remove(_FILENAME)

        @it.should("parse a file without errors")
        def test():
            it.source = VhdlParser(Path(_FILENAME))

        @it.should("return its entities")  # type: ignore
        def test():
            design_units = list(it.source.getDesignUnits())
            _logger.info("Design units: %s", design_units)
            it.assertNotEqual(design_units, None, "No design_units units found")
            it.assertCountEqual(
                design_units,
                [
                    VhdlDesignUnit(
                        owner=it.source.filename,
                        type_=DesignUnitType.entity,
                        name="clock_divider",
                        locations={(14, 7)},
                    )
                ],
            )

        @it.should("parse its libraries")  # type: ignore
        def test():
            libraries = it.source.getLibraries()
            _logger.info("Libraries found: %s", ", ".join([repr(x) for x in libraries]))

            it.assertCountEqual(
                libraries, ["ieee", "lib1", "lib2", "lib3", "lib4", "lib5", "lib6"]
            )

        @it.should("return its dependencies")  # type: ignore
        def test():
            it.assertCountEqual(
                it.source.getDependencies(),
                [
                    _DependencySpec(
                        owner=Path(_FILENAME),
                        library="ieee",
                        name="std_logic_1164",
                        locations=[(1, 4)],
                    ),
                    _DependencySpec(
                        owner=Path(_FILENAME),
                        library="ieee",
                        name="std_logic_arith",
                        locations=[(2, 5)],
                    ),
                    _DependencySpec(
                        owner=Path(_FILENAME),
                        library=None,
                        name="package_with_constants",
                        locations=[(5, 5)],
                    ),
                    _DependencySpec(
                        owner=Path(_FILENAME),
                        library=None,
                        name="cherry_pick",
                        locations=[(6, 6), (7, 4)],
                    ),
                ],
            )

        @it.should("return source modification time")  # type: ignore
        def test():
            it.assertEqual(os.path.getmtime(_FILENAME), it.source.getmtime())

        @it.should("return updated dependencies")  # type: ignore
        def test():
            code = list(it._code)

            code.insert(0, "library some_library;")
            code.insert(1, "    use some_library.some_package;")
            writeListToFile(_FILENAME, code)

            time.sleep(0.1)

            it.assertCountEqual(
                it.source.getDependencies(),
                [
                    _DependencySpec(
                        owner=Path(_FILENAME),
                        library="ieee",
                        name="std_logic_1164",
                        locations=[(3, 4)],
                    ),
                    _DependencySpec(
                        owner=Path(_FILENAME),
                        library="ieee",
                        name="std_logic_arith",
                        locations=[(4, 5)],
                    ),
                    _DependencySpec(
                        owner=Path(_FILENAME),
                        library=None,
                        name="package_with_constants",
                        locations=[(7, 5)],
                    ),
                    _DependencySpec(
                        owner=Path(_FILENAME),
                        library="some_library",
                        name="some_package",
                        locations=[(1, 8)],
                    ),
                    _DependencySpec(
                        owner=Path(_FILENAME),
                        library=None,
                        name="cherry_pick",
                        locations=[(8, 6), (9, 4)],
                    ),
                ],
            )

        @it.should("handle implicit libraries")  # type: ignore
        def test():
            code = list(it._code)
            code.insert(0, "    use work.another_package;")
            writeListToFile(_FILENAME, code)

            it.assertCountEqual(
                it.source.getDependencies(),
                [
                    _DependencySpec(
                        owner=Path(_FILENAME),
                        library="ieee",
                        name="std_logic_1164",
                        locations=[(2, 4)],
                    ),
                    _DependencySpec(
                        owner=Path(_FILENAME),
                        library="ieee",
                        name="std_logic_arith",
                        locations=[(3, 5)],
                    ),
                    _DependencySpec(
                        owner=Path(_FILENAME),
                        library=None,
                        name="package_with_constants",
                        locations=[(6, 5)],
                    ),
                    _DependencySpec(
                        owner=Path(_FILENAME),
                        library=None,
                        name="another_package",
                        locations=[(0, 8)],
                    ),
                    _DependencySpec(
                        owner=Path(_FILENAME),
                        library=None,
                        name="cherry_pick",
                        locations=[(7, 6), (8, 4)],
                    ),
                ],
            )

        @it.should("handle libraries without packages")  # type: ignore
        def test():
            code = list(it._code)
            code.insert(0, "library remove_me;")
            writeListToFile(_FILENAME, code)

            it.assertCountEqual(
                it.source.getDependencies(),
                [
                    _DependencySpec(
                        owner=Path(_FILENAME),
                        library="ieee",
                        name="std_logic_1164",
                        locations=[(2, 4)],
                    ),
                    _DependencySpec(
                        owner=Path(_FILENAME),
                        library="ieee",
                        name="std_logic_arith",
                        locations=[(3, 5)],
                    ),
                    _DependencySpec(
                        owner=Path(_FILENAME),
                        library=None,
                        name="package_with_constants",
                        locations=[(6, 5)],
                    ),
                    _DependencySpec(
                        owner=Path(_FILENAME),
                        library=None,
                        name="cherry_pick",
                        locations=[(7, 6), (8, 4)],
                    ),
                ],
            )

        @it.should(  # type: ignore
            "report as equal after recovering from cache via json"
        )
        def test():
            state = json.dumps(it.source, cls=StateEncoder)
            _logger.info("State before: %s", state)
            recovered = json.loads(state, object_hook=jsonObjectHook)
            it.assertEqual(it.source.filename, recovered.filename)

    with it.having("a package code"):

        @it.has_setup
        def setup():
            it._code = [
                "library ieee;",
                "use ieee.std_logic_1164.all;",
                "use ieee.std_logic_arith.all;",
                "use ieee.std_logic_unsigned.all;",
                "",
                "library basic_library;",
                "",
                "package package_with_constants is",
                "",
                "    constant SOME_INTEGER_CONSTANT : integer := 10;",
                '    constant SOME_STRING_CONSTANT  : string := "Hello";',
                "    constant ANOTHER_CONSTANT : string := work.foo.bar;",
                "",
                "    constant SOME_STRING : string := "
                "basic_library.very_common_pkg.VIM_HDL_VERSION;",
                "end;",
                "",
                "package body package_with_constants is",
                "",
                "end package body;",
                "",
                "package body package_body_only is",
                "",
                "end package body package_body_only;",
            ]

            writeListToFile(_FILENAME, it._code)
            it._source_mtime = os.path.getmtime(_FILENAME)

        @it.has_teardown
        def teardown():
            if os.path.exists(_FILENAME):
                os.remove(_FILENAME)

        @it.should("parse a file without errors")  # type: ignore
        def test():
            it.source = VhdlParser(Path(_FILENAME))

        @it.should("return the names of the packages found")  # type: ignore
        def test():
            it.assertCountEqual(
                list(it.source.getDesignUnits()),
                [
                    VhdlDesignUnit(
                        owner=it.source.filename,
                        type_=DesignUnitType.package,
                        name="package_with_constants",
                        locations={(7, 8)},
                    )
                ],
            )

        @it.should("return its dependencies")  # type: ignore
        def test():  # type: () -> None
            it.assertCountEqual(
                it.source.getDependencies(),
                [
                    _DependencySpec(
                        name="std_logic_1164",
                        library="ieee",
                        owner=it.source.filename,
                        locations={(1, 4)},
                    ),
                    _DependencySpec(
                        name="std_logic_arith",
                        library="ieee",
                        owner=it.source.filename,
                        locations={(2, 4)},
                    ),
                    _DependencySpec(
                        name="std_logic_unsigned",
                        library="ieee",
                        owner=it.source.filename,
                        locations={(3, 4)},
                    ),
                    _DependencySpec(
                        name="foo",
                        library=None,
                        owner=it.source.filename,
                        locations={(11, 42)},
                    ),
                    _DependencySpec(
                        name="very_common_pkg",
                        library="basic_library",
                        owner=it.source.filename,
                        locations={(13, 37)},
                    ),
                    _DependencySpec(
                        name="package_with_constants",
                        library=None,
                        owner=it.source.filename,
                        locations={(16, 0)},
                    ),
                    _DependencySpec(
                        name="package_body_only",
                        library=None,
                        owner=it.source.filename,
                        locations={(20, 0)},
                    ),
                ],
            )

        @it.should("return source modification time")  # type: ignore
        def test():
            it.assertEqual(os.path.getmtime(_FILENAME), it.source.getmtime())

    with it.having("a context code"):

        @it.has_setup
        def setup():
            it._code = [
                "context context_name is",
                "  library lib0;",
                "  use lib0.pkg0.all;",
                "  use lib0.pkg1.name1;",
                "  library lib1;",
                "  use lib1.all;",
                "end context;",
            ]

            writeListToFile(_FILENAME, it._code)
            it._source_mtime = os.path.getmtime(_FILENAME)

        @it.has_teardown
        def teardown():
            if os.path.exists(_FILENAME):
                os.remove(_FILENAME)

        @it.should("create the object with no errors")  # type: ignore
        def test():
            it.source = VhdlParser(Path(_FILENAME))

        @it.should("return the names of the packages found")  # type: ignore
        def test():
            it.assertCountEqual(
                list(it.source.getDesignUnits()),
                [
                    VhdlDesignUnit(
                        owner=it.source.filename,
                        type_=DesignUnitType.context,
                        name="context_name",
                        locations={(0, 7)},
                    )
                ],
            )

        @it.should("return its dependencies")  # type: ignore
        def test():  # type: () -> None
            it.assertCountEqual(
                it.source.getDependencies(),
                [
                    _DependencySpec(
                        owner=it.source.filename,
                        name="pkg0",
                        library="lib0",
                        locations={(2, 6)},
                    ),
                    _DependencySpec(
                        owner=it.source.filename,
                        name="pkg1",
                        library="lib0",
                        locations={(3, 6)},
                    ),
                    _DependencySpec(
                        owner=it.source.filename,
                        name="all",
                        library="lib1",
                        locations={(5, 6)},
                    ),
                ],
            )

        @it.should("return source modification time")  # type: ignore
        def test():
            it.assertEqual(os.path.getmtime(_FILENAME), it.source.getmtime())


it.createTests(globals())



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Language server protocol implementation"

import logging
import os.path as p
from threading import Timer
from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Tuple, Union

import six
from pyls import lsp as defines  # type: ignore
from pyls._utils import debounce  # type: ignore
from pyls.python_ls import PythonLanguageServer  # type: ignore
from pyls.uris import from_fs_path, to_fs_path  # type: ignore
from pyls.workspace import Workspace  # type: ignore
from tabulate import tabulate

from hdl_checker import DEFAULT_PROJECT_FILE
from hdl_checker.base_server import HdlCodeCheckerBase
from hdl_checker.config_generators.simple_finder import SimpleFinder
from hdl_checker.database import _DEFAULT_LIBRARY_NAME
from hdl_checker.diagnostics import CheckerDiagnostic, DiagType
from hdl_checker.exceptions import HdlCheckerNotInitializedError, UnknownParameterError
from hdl_checker.parsers.elements.dependency_spec import DependencySpec
from hdl_checker.parsers.elements.design_unit import (
    VerilogDesignUnit,
    VhdlDesignUnit,
    tAnyDesignUnit,
)
from hdl_checker.path import Path
from hdl_checker.types import Location, MarkupKind
from hdl_checker.utils import logCalls

_logger = logging.getLogger(__name__)

LINT_DEBOUNCE_S = 0.5  # 500 ms

URI = str

_SETTING_UP_A_PROJECT_URL = (
    "https://github.com/suoto/hdl_checker/wiki/Setting-up-a-project"
)

_SETUP_IS_TOO_LONG_TIMEOUT = 15

_SETUP_IS_TOO_LONG_MSG = (
    "Configuring the project seems to be taking too long. Consider using a "
    "smaller workspace or a configuration file. More info: "
    "[{0}]({0})".format(_SETTING_UP_A_PROJECT_URL)
)

if six.PY2:
    FileNotFoundError = (  # pylint: disable=redefined-builtin,invalid-name
        IOError,
        OSError,
    )


def checkerDiagToLspDict(diag):
    # type: (...) -> Any
    """
    Converts a CheckerDiagnostic object into the dictionary with into the LSP
    expects
    """
    _logger.debug(diag)

    # Translate the error into LSP severity
    severity = diag.severity

    if severity in (DiagType.INFO, DiagType.STYLE_INFO):
        severity = defines.DiagnosticSeverity.Hint
    elif severity in (DiagType.STYLE_WARNING, DiagType.STYLE_ERROR):
        severity = defines.DiagnosticSeverity.Information
    elif severity in (DiagType.WARNING,):
        severity = defines.DiagnosticSeverity.Warning
    elif severity in (DiagType.ERROR,):
        severity = defines.DiagnosticSeverity.Error
    else:
        severity = defines.DiagnosticSeverity.Error

    result = {
        "source": diag.checker,
        "range": {
            "start": {
                "line": (diag.line_number or 1) - 1,
                "character": (diag.column_number or 1) - 1,
            },
            "end": {
                "line": (diag.line_number or 1) - 1,
                "character": (diag.column_number or 1) - 1,
            },
        },
        "message": diag.text,
        "severity": severity,
    }

    if diag.error_code:
        result["code"] = diag.error_code

    return result


class HdlCodeCheckerServer(HdlCodeCheckerBase):
    """
    HDL Checker project builder class
    """

    def __init__(self, workspace, root_path):
        # type: (Workspace, str) -> None
        self._workspace = workspace
        super(HdlCodeCheckerServer, self).__init__(Path(root_path))

    def _handleUiInfo(self, message):
        # type: (...) -> Any
        _logger.debug("UI info: %s (workspace=%s)", message, self._workspace)
        if self._workspace:  # pragma: no cover
            self._workspace.show_message(message, defines.MessageType.Info)

    def _handleUiWarning(self, message):
        # type: (...) -> Any
        _logger.debug("UI warning: %s (workspace=%s)", message, self._workspace)
        if self._workspace:  # pragma: no cover
            self._workspace.show_message(message, defines.MessageType.Warning)

    def _handleUiError(self, message):
        # type: (...) -> Any
        _logger.debug("UI error: %s (workspace=%s)", message, self._workspace)
        if self._workspace:  # pragma: no cover
            self._workspace.show_message(message, defines.MessageType.Error)


class HdlCheckerLanguageServer(PythonLanguageServer):
    """ Implementation of the Microsoft VSCode Language Server Protocol
    https://github.com/Microsoft/language-server-protocol/blob/master/versions/protocol-1-x.md
    """

    # pylint: disable=too-many-public-methods,redefined-builtin

    def __init__(self, *args, **kwargs):
        # type: (...) -> None
        self._checker = None  # type: Optional[HdlCodeCheckerServer]
        super(HdlCheckerLanguageServer, self).__init__(*args, **kwargs)
        # Default checker
        self._onConfigUpdate({"project_file": None})
        self._global_diags = set()  # type: Set[CheckerDiagnostic]
        self._initialization_options = {}  # type: Dict[str, Any]

    @property
    def checker(self):
        # type: () -> HdlCodeCheckerServer
        """
        Returns a valid checker, otherwise raise HdlCheckerNotInitializedError.
        In theory this shouldn't happen since root URI is mandatory when
        initializing the server.
        """
        if self._checker is None:
            _logger.error("Checker not initialized")
            raise HdlCheckerNotInitializedError()
        return self._checker

    def showInfo(self, msg):
        # type: (str) -> None
        """
        Shorthand for self.workspace.show_message(msg, defines.MessageType.Info)
        """
        self.workspace.show_message(msg, defines.MessageType.Info)

    def showWarning(self, msg):
        # type: (str) -> None
        """
        Shorthand for self.workspace.show_message(msg, defines.MessageType.Warning)
        """
        self.workspace.show_message(msg, defines.MessageType.Warning)

    def capabilities(self):
        # type: (...) -> Any
        "Returns language server capabilities"
        return {
            "definitionProvider": True,
            "hoverProvider": True,
            "textDocumentSync": defines.TextDocumentSyncKind.FULL,
        }

    @logCalls
    def m_initialized(self, **_kwargs):
        """
        Enables processing of actions that were generated upon m_initialize and
        were delayed because the client might need further info (for example to
        handle window/showMessage requests)
        """
        self._onConfigUpdate(self._initialization_options)
        return super(HdlCheckerLanguageServer, self).m_initialized(**_kwargs)

    @logCalls
    def m_initialize(
        self,
        processId=None,
        rootUri=None,
        rootPath=None,
        initializationOptions=None,
        **_kwargs
    ):
        # type: (...) -> Any
        """
        Initializes the language server
        """
        result = super(HdlCheckerLanguageServer, self).m_initialize(
            processId=processId,
            rootUri=rootUri,
            rootPath=rootPath,
            initializationOptions={},
            **_kwargs
        )
        self._initialization_options = initializationOptions
        return result

    def _onConfigUpdate(self, options):
        # type: (...) -> Any
        """
        Updates the checker server from options if the 'project_file' key is
        present. Please not that this is run from both initialize and
        workspace/did_change_configuration and when ran initialize the LSP
        client might not ready to take messages. To circumvent this, make sure
        m_initialize returns before calling this to actually configure the
        server.
        """
        if not self.workspace or not self.workspace.root_uri:
            return

        root_path = to_fs_path(self.workspace.root_uri)
        self._checker = HdlCodeCheckerServer(self.workspace, root_path=root_path)

        _logger.debug("Updating from %s", options)

        # Clear previous diagnostics
        self._global_diags = set()

        path = self._getProjectFilePath(options)

        try:
            self.checker.setConfig(path)
            return
        except UnknownParameterError as exc:
            _logger.info("Failed to read config from %s: %s", path, exc)
            return
        except FileNotFoundError:
            # If the file couldn't be found, proceed to searching the root
            # URI (if it has been set)
            pass

        if not self.workspace or not self.workspace.root_path:
            _logger.debug("No workspace and/or root path not set, can't search files")
            return

        # Don't let the user hanging if adding sources is taking too long. Also
        # need to notify when adding is done.
        timer = Timer(
            _SETUP_IS_TOO_LONG_TIMEOUT, self.showWarning, args=(_SETUP_IS_TOO_LONG_MSG,)
        )
        timer.start()

        self.showInfo("Searching {} for HDL files...".format(self.workspace.root_path))

        # Having no project file but with root URI triggers searching for
        # sources automatically
        config = SimpleFinder([self.workspace.root_path]).generate()
        source_cnt = len(config["sources"])  # key will be consumed by the checker
        self.checker.configure(config)

        timer.cancel()
        self.showInfo("Added {} sources".format(source_cnt))

    def _getProjectFilePath(self, options=None):
        # type: (...) -> str
        """
        Tries to get 'project_file' from the options dict and combine it with
        the root URI as provided by the workspace
        """
        path = (options or {}).get("project_file", DEFAULT_PROJECT_FILE)

        # Project file will be related to the root path
        if self.workspace:
            path = p.join(self.workspace.root_path, path)

        return path

    @debounce(LINT_DEBOUNCE_S, keyed_by="doc_uri")
    def lint(self, doc_uri, is_saved):
        # type: (...) -> Any
        diagnostics = set(self._getDiags(doc_uri, is_saved)) | self._global_diags

        # Since we're debounced, the document may no longer be open
        if doc_uri in self.workspace.documents:
            # Both checker methods return generators, convert to a list before
            # returning
            self.workspace.publish_diagnostics(
                doc_uri, list([checkerDiagToLspDict(x) for x in diagnostics])
            )

    def _getDiags(self, doc_uri, is_saved):
        # type: (URI, bool) -> Iterable[CheckerDiagnostic]
        """
        Gets diags of the URI, wether from the saved file or from its
        contents
        """
        if self.checker is None:  # pragma: no cover
            _logger.debug("No checker, won't try to get diagnostics")
            return ()

        # If the file has not been saved, use the appropriate method, which
        # will involve dumping the modified contents into a temporary file
        path = Path(to_fs_path(doc_uri))

        _logger.info("Linting %s (saved=%s)", repr(path), is_saved)

        if is_saved:
            diags = self.checker.getMessagesByPath(path)
        else:
            text = self.workspace.get_document(doc_uri).source
            diags = self.checker.getMessagesWithText(path, text)

        # LSP diagnostics are only valid for the scope of the resource and
        # hdl_checker may return a tree of issues, so need to filter those out
        return (diag for diag in diags if diag.filename in (path, None))

    def m_workspace__did_change_configuration(self, settings=None):
        # type: (...) -> Any
        self._onConfigUpdate(settings or {})

    @property
    def _use_markdown_for_hover(self):
        """
        Returns True if the client has reported 'markdown' as one of the
        supported formats, i.e., 'markdown' is present inside
        TextDocumentClientCapabilities.hover.contentFormat
        """
        return MarkupKind.Markdown.value in (
            self.config.capabilities.get("textDocument", {})
            .get("hover", {})
            .get("contentFormat", [])
        )

    def _format(self, text):
        """
        Double line breaks if workspace supports markdown
        """
        if self._use_markdown_for_hover:
            return text.replace("\n", "\n\n")

        return text

    def _getBuildSequenceForHover(self, path):
        # type: (Path) -> str
        """
        Return a formatted text with the build sequence for the given path
        """
        sequence = []  # type: List[Tuple[int, str, str]]

        # Adds the sequence of dependencies' paths
        for i, (seq_library, seq_path) in enumerate(
            self.checker.database.getBuildSequence(
                path, self.checker.builder.builtin_libraries
            ),
            1,
        ):
            sequence += [(i, str(seq_library), str(seq_path))]

        # Adds the original path
        sequence += [
            (
                len(sequence) + 1,
                str(self.checker.database.getLibrary(path) or _DEFAULT_LIBRARY_NAME),
                str(path),
            )
        ]

        return "Build sequence for {} is\n\n{}".format(
            path,
            tabulate(
                sequence,
                tablefmt="github" if self._use_markdown_for_hover else "plain",
                headers=("#", "Library", "Path"),
            ),
        )

    def _getDependencyInfoForHover(self, dependency):
        # type: (DependencySpec) -> Optional[str]
        """
        Report which source defines a given dependency when the user hovers
        over its name
        """
        # If that doesn't match, check for dependencies
        info = self.checker.resolveDependencyToPath(dependency)
        if info is not None:
            return self._format('Path "{}", library "{}"'.format(info[0], info[1]))

        return "Couldn't find a source defining '{}.{}'".format(
            dependency.library, dependency.name
        )

    def _getElementAtPosition(self, path, position):
        # type: (Path, Location) -> Union[DependencySpec, tAnyDesignUnit, None]
        """
        Gets design units and dependencies (in this order) of path and checks
        if their definitions include position
        """
        for meth in (
            self.checker.database.getDesignUnitsByPath,
            self.checker.database.getDependenciesByPath,
        ):  # type: Callable
            for element in meth(path):
                if element.includes(position):
                    return element

        return None

    def hover(self, doc_uri, position):
        # type: (URI, Dict[str, int]) -> Any
        path = Path(to_fs_path(doc_uri))
        # Check if the element under the cursor matches something we know
        element = self._getElementAtPosition(
            path, Location(line=position["line"], column=position["character"])
        )

        _logger.debug("Getting info from %s", element)

        if isinstance(element, (VerilogDesignUnit, VhdlDesignUnit)):
            return {"contents": self._getBuildSequenceForHover(path)}

        if isinstance(element, DependencySpec):
            return {"contents": self._getDependencyInfoForHover(element)}

        return {}

    @logCalls
    def definitions(self, doc_uri, position):
        # type: (...) -> Any
        doc_path = Path(to_fs_path(doc_uri))
        dependency = self._getElementAtPosition(
            doc_path, Location(line=position["line"], column=position["character"])
        )

        if not isinstance(dependency, DependencySpec):
            _logger.debug("Go to definition not supported for item %s", dependency)
            return []

        # Work out where this dependency refers to
        info = self.checker.resolveDependencyToPath(dependency)

        if info is None:
            _logger.debug("Unable to resolve %s to a path", dependency)
            return []

        _logger.info("Dependency %s resolved to %s", dependency, info)

        # Make the response
        target_path, _ = info
        target_uri = from_fs_path(str(target_path))

        locations = []  # type: List[Dict[str, Any]]

        # Get the design unit that has matched the dependency to extract the
        # location where it's defined
        for unit in self.checker.database.getDesignUnitsByPath(target_path):
            if unit.name == dependency.name and unit.locations:
                for line, column in unit.locations:
                    locations += [
                        {
                            "uri": target_uri,
                            "range": {
                                "start": {"line": line, "character": column},
                                "end": {"line": line, "character": column + len(unit)},
                            },
                        }
                    ]

        return locations



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"HDL Checker project builder class"

import abc
import json
import logging
import os
import os.path as p
import tempfile
import traceback
from collections import namedtuple
from multiprocessing.pool import ThreadPool
from pprint import pformat
from threading import RLock
from typing import Any, AnyStr, Dict, Iterable, Optional, Set, Tuple, Union

import six

from hdl_checker import CACHE_NAME, WORK_PATH
from hdl_checker.builder_utils import (
    getBuilderByName,
    getVunitSources,
    getWorkingBuilders,
)
from hdl_checker.builders.fallback import Fallback
from hdl_checker.database import Database
from hdl_checker.diagnostics import CheckerDiagnostic, DiagType, PathNotInProjectFile
from hdl_checker.parsers.config_parser import ConfigParser
from hdl_checker.parsers.elements.dependency_spec import DependencySpec
from hdl_checker.parsers.elements.identifier import Identifier
from hdl_checker.path import Path, TemporaryPath
from hdl_checker.serialization import StateEncoder, jsonObjectHook
from hdl_checker.static_check import getStaticMessages
from hdl_checker.types import (
    BuildFlagScope,
    RebuildInfo,
    RebuildLibraryUnit,
    RebuildPath,
    RebuildUnit,
)
from hdl_checker.utils import removeDirIfExists, removeIfExists, toBytes

try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache  # type: ignore

_logger = logging.getLogger(__name__)

if six.PY2:
    JSONDecodeError = ValueError
else:
    JSONDecodeError = json.decoder.JSONDecodeError

WatchedFile = namedtuple("WatchedFile", ("path", "last_read"))


class HdlCodeCheckerBase(object):  # pylint: disable=useless-object-inheritance
    """
    HDL Checker project builder class
    """

    _USE_THREADS = True
    _MAX_REBUILD_ATTEMPTS = 20

    __metaclass__ = abc.ABCMeta

    def __init__(self, root_dir):  # type: (Path) -> None
        # Root dir is the absolute path to use when any path passed on is
        # relative
        self.root_dir = root_dir
        # Work dir is the the directory that HDL Checker uses as a scratch
        # pad, everything within it may be deleted or changed
        self.work_dir = Path(p.join(str(self.root_dir), WORK_PATH))

        self._lock = RLock()
        self.config_file = None  # type: Optional[WatchedFile]

        self._database = Database()
        self._builder = Fallback(self.work_dir, self.database)

        if not p.exists(str(self.work_dir)):
            os.makedirs(str(self.work_dir))

        self._recoverCacheIfPossible()
        self._saveCache()

        # Use this to know which methods should be cache
        self._cached_methods = {
            getattr(self, x)
            for x in dir(self)
            if hasattr(getattr(self, x), "cache_clear")
        }

    @property
    def builder(self):
        """
        Parses the config file if it has been set and returns the builder in
        use
        """
        self._updateConfigIfNeeded()
        return self._builder

    @property
    def database(self):
        """
        Parses the config file if it has been set and returns the builder in
        use
        """
        self._updateConfigIfNeeded()
        return self._database

    def __hash__(self):
        # Just to allow lru_cache
        return hash(0)

    def _clearLruCaches(self):
        "Clear caches from lru_caches"
        for meth in self._cached_methods:
            meth.cache_clear()

    def setConfig(self, filename):
        # type: (Union[Path, str]) -> None
        """
        Sets the configuration file. Calling this method will only trigger a
        configuration update if the given filename is different what was
        configured previously (that includes no value previously set)
        """
        path = Path(filename, self.root_dir)
        mtime = path.mtime

        # If the config file has been set previously, avoid refreshing if
        # possible
        if self.config_file is None or self.config_file.path != path:
            _logger.info("Replacing %s with %s", self.config_file, path)
            mtime = 0.0
        else:
            return

        self.config_file = WatchedFile(path, mtime)
        _logger.debug("Set config to %s", self.config_file)

    def _updateConfigIfNeeded(self):
        # type: (...) -> Any
        """
        Checks if self.config_file has changed; if it has, cleans up and
        re-reads the file
        """
        # No config file set
        if self.config_file is None:
            return

        file_mtime = self.config_file.path.mtime
        # Check if values we have are up to date
        if self.config_file.last_read >= file_mtime:
            return

        self.clean()
        self._readConfig()

    def _readConfig(self):
        # type: (...) -> None
        """
        Updates the database from a configuration file. Extracting form JSON
        has priority, then ConfigParser is attempted.
        """
        if self.config_file is None:
            _logger.warning("Can't read config when config file is not set")
            return

        try:
            config = json.load(open(str(self.config_file.path)))
        except JSONDecodeError:
            config = ConfigParser(self.config_file.path).parse()

        self.config_file = WatchedFile(
            self.config_file.path, self.config_file.path.mtime
        )
        _logger.debug("Updating config file to %s", self.config_file)
        self.configure(config)

    def configure(self, config):
        # type: (Dict[Any, Any]) -> None
        "Updates configuration from a dictionary"

        _logger.debug("Updating with base config:\n%s", pformat(config))

        builder_name = config.pop("builder", None)
        if builder_name is not None:
            builder_cls = getBuilderByName(builder_name)
        else:
            try:
                builder_cls = list(getWorkingBuilders()).pop()
            except IndexError:
                builder_cls = Fallback

        _logger.debug("Builder class: %s", builder_cls)

        self._builder = builder_cls(self.work_dir, self.database)

        self.database.configure(config, str(self.root_dir))
        # Add VUnit
        if not isinstance(self._builder, Fallback):
            for path, library, flags in getVunitSources(self._builder):
                self.database.addSource(path, library, flags, flags)

        # Add the flags from the root config file last, it should overwrite
        # values set by the included files
        if config:
            _logger.warning(
                "Some configuration elements weren't used:\n%s", pformat(config)
            )

    def _getCacheFilename(self):
        # type: () -> Path
        """
        The cache file name will always be inside the path returned by self._getWorkingPath
        and defaults to cache.json
        """
        return Path(CACHE_NAME, self.work_dir)

    def _saveCache(self):
        # type: (...) -> Any
        """
        Dumps project object to a file to recover its state later
        """
        cache_fname = self._getCacheFilename()

        state = {
            "builder": self.builder,
            "config_file": self.config_file,
            "database": self.database,
        }

        _logger.debug("Saving state to '%s'", cache_fname)
        if not p.exists(p.dirname(cache_fname.name)):
            os.makedirs(p.dirname(cache_fname.name))
        json.dump(state, open(cache_fname.name, "w"), indent=True, cls=StateEncoder)

    def _setState(self, state):
        # type: (...) -> Any
        """
        Serializer load implementation
        """
        self._database = state.pop("database")
        self._builder = state.pop("builder", Fallback)
        config_file = state.pop("config_file", None)
        if config_file is None:
            self.config_file = None
        else:
            WatchedFile._make(config_file)

    def _recoverCacheIfPossible(self):
        # type: (...) -> Any
        """
        Tries to recover cached info for the given config_file. If
        something goes wrong, assume the cache is invalid and return
        nothing. Otherwise, return the cached object
        """
        cache_fname = self._getCacheFilename()

        try:
            cache = json.load(open(cache_fname.name, "r"), object_hook=jsonObjectHook)
            _logger.debug("Recovered cache from '%s'", cache_fname)
        except IOError:
            _logger.debug("Couldn't read cache file %s, skipping recovery", cache_fname)
            return
        except ValueError as exception:
            self._handleUiWarning(
                "Unable to recover cache from '{}': {}".format(
                    cache_fname, str(exception)
                )
            )

            _logger.warning(
                "Unable to recover cache from '%s': %s",
                cache_fname,
                traceback.format_exc(),
            )
            return

        self._setState(cache)
        self._builder.setup()

    def _cleanIfNeeded(self):
        # type: (...) -> Any
        """
        Sanity checks to make sure the environment is sane
        """
        if not p.exists(self._builder.work_folder):
            self._builder.setup()

    def clean(self):
        # type: (...) -> Any
        """
        Clean up generated files
        """
        _logger.debug("Cleaning up project")
        removeDirIfExists(str(self.work_dir))

        del self._builder
        del self._database

        database = Database()
        self._database = database
        self._builder = Fallback(self.work_dir, database)

    @abc.abstractmethod
    def _handleUiInfo(self, message):  # type: (AnyStr) -> None
        """
        Method that should be overriden to handle info messages from
        HDL Checker to the user
        """

    @abc.abstractmethod
    def _handleUiWarning(self, message):  # type: (AnyStr) -> None
        """
        Method that should be overriden to handle warning messages
        from HDL Checker to the user
        """

    @abc.abstractmethod
    def _handleUiError(self, message):  # type: (AnyStr) -> None
        """
        Method that should be overriden to handle errors messages
        from HDL Checker to the user
        """

    def _getBuilderMessages(self, path):
        # type: (Path) -> Iterable[CheckerDiagnostic]
        """
        Builds the given path taking care of recursively building its
        dependencies first
        """
        _logger.debug("Building '%s'", str(path))

        path = Path(path, self.root_dir)

        for dep_library, dep_path in self.database.getBuildSequence(
            path, self.builder.builtin_libraries
        ):
            for record in self._buildAndHandleRebuilds(
                dep_path, dep_library, scope=BuildFlagScope.dependencies
            ):
                if record.severity in (DiagType.ERROR, DiagType.STYLE_ERROR):
                    yield record

        _logger.debug("Built dependencies, now actually building '%s'", str(path))
        library = self.database.getLibrary(path)
        for record in self._buildAndHandleRebuilds(
            path,
            library if library is not None else Identifier("work"),
            scope=BuildFlagScope.single,
            forced=True,
        ):
            yield record

    def _buildAndHandleRebuilds(self, path, library, scope, forced=False):
        # type: (Path, Identifier, BuildFlagScope, bool) -> Iterable[CheckerDiagnostic]
        """
        Builds the given path and handle any files that might require
        rebuilding until there is nothing to rebuild. The number of iteractions
        is fixed in 10.
        """
        self._cleanIfNeeded()
        # Limit the amount of calls to rebuild the same file to avoid
        # hanging the server
        for _ in range(self._MAX_REBUILD_ATTEMPTS):
            records, rebuilds = self.builder.build(
                path=path, library=library, scope=scope, forced=forced
            )

            if rebuilds:
                _logger.debug(
                    "Building '%s' triggers rebuilding: %s",
                    path,
                    ", ".join([str(x) for x in rebuilds]),
                )
                self._handleRebuilds(rebuilds)
            else:
                _logger.debug("Had no rebuilds for %s", path)
                return records

        self._handleUiError(
            "Unable to build '%s' after %d attempts"
            % (path, self._MAX_REBUILD_ATTEMPTS)
        )

        return {}

    def _handleRebuilds(self, rebuilds):
        # type: (Iterable[RebuildInfo]) -> None
        """
        Resolves hints found in the rebuild list into path objects
        and rebuild them
        """
        for rebuild in rebuilds:
            _logger.debug("Rebuild hint: '%s'", rebuild)
            if isinstance(rebuild, RebuildUnit):
                for path in self.database.getPathsDefining(name=rebuild.name):
                    list(self._getBuilderMessages(path))

            elif isinstance(rebuild, RebuildLibraryUnit):
                for path in self.database.getPathsDefining(
                    name=rebuild.name, library=rebuild.library
                ):
                    list(self._getBuilderMessages(path))
            elif isinstance(rebuild, RebuildPath):
                list(self._getBuilderMessages(rebuild.path))

            else:  # pragma: no cover
                _logger.warning("Did nothing with %s", rebuild)

    def getMessagesByPath(self, path):
        # type: (Path) -> Iterable[CheckerDiagnostic]
        """
        Returns the messages for the given path, including messages
        from the configured builder (if available) and static checks
        """
        self._clearLruCaches()

        path = Path(path, self.root_dir)

        builder_diags = set()  # type: Set[CheckerDiagnostic]

        if self._USE_THREADS:
            pool = ThreadPool()

            static_check = pool.apply_async(
                getStaticMessages, args=(tuple(open(path.name).read().split("\n")),)
            )

            builder_check = pool.apply_async(self._getBuilderMessages, args=[path])
            builder_diags |= {x for x in builder_check.get()}

            pool.close()
            pool.join()

            static_diags = {x for x in static_check.get()}

        else:  # pragma: no cover
            builder_diags |= set(self._getBuilderMessages(path))
            static_diags = set(
                getStaticMessages(tuple(open(path.name).read().split("\n")))
            )

        # Static messages don't take the path, only the text, so we need to
        # set add that to the diagnostic
        for diag in static_diags:
            diag.filename = path
            builder_diags.add(diag)

        self._saveCache()

        diags = builder_diags | set(self.database.getDiagnosticsForPath(path))

        # If we're working off of a project file, no need to filter out
        # diags about path not being found
        if self.config_file is not None:
            return diags

        return {diag for diag in diags if not isinstance(diag, PathNotInProjectFile)}

    def getMessagesWithText(self, path, content):
        # type: (Path, AnyStr) -> Iterable[CheckerDiagnostic]
        """
        Dumps content to a temprary file and replaces the temporary file name
        for path on the diagnostics received
        """
        self._clearLruCaches()
        with self._lock:
            _logger.info("Getting messages for '%s' with content", path)

            ext = path.name.split(".")[-1]
            temporary_file = tempfile.NamedTemporaryFile(suffix="." + ext, delete=False)

            temp_path = TemporaryPath(temporary_file.name)

            temporary_file.file.write(toBytes(content))  # type: ignore
            temporary_file.close()

            # If the reference path was added to the database, add the
            # temporary file with the same attributes
            if path in self.database.paths:
                library = self.database.getLibrary(path)
                self.database.addSource(
                    temp_path,
                    getattr(library, "display_name", None),
                    self.database.getFlags(path, BuildFlagScope.single),
                    self.database.getFlags(path, BuildFlagScope.dependencies),
                )

            diags = set()  # type: Set[CheckerDiagnostic]

            # Some messages may not include the filename field when checking a
            # file by content. In this case, we'll assume the empty filenames
            # refer to the same filename we got in the first place
            for diag in self.getMessagesByPath(temp_path):
                diag.filename = path
                diag.text = diag.text.replace(temporary_file.name, path.name)
                diags.add(diag)

            self.database.removeSource(temp_path)
            removeIfExists(temporary_file.name)

            diags |= set(self.database.getDiagnosticsForPath(path))

            if self.config_file and path not in self.database.paths:
                diags.add(PathNotInProjectFile(path))

        return diags

    @lru_cache()
    def resolveDependencyToPath(self, dependency):
        # type: (DependencySpec) -> Optional[Tuple[Path, Identifier]]
        """
        Retrieves the build sequence for the dependency's owner and extracts
        the path that implements a design unit whose names match the
        dependency.
        """
        for library, path in self.database.getBuildSequence(
            dependency.owner, self.builder.builtin_libraries
        ):
            if dependency.name in (
                x.name for x in self.database.getDesignUnitsByPath(path)
            ):
                return path, library

        return None



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=function-redefined
# pylint: disable=missing-docstring
# pylint: disable=protected-access
# pylint: disable=useless-object-inheritance

import json
import logging
import os
import os.path as p
import shutil
import time

import six
from mock import patch

from nose2.tools import such  # type: ignore

from hdl_checker.tests import (
    FailingBuilder,
    MockBuilder,
    PatchBuilder,
    SourceMock,
    StandaloneProjectBuilder,
    assertCountEqual,
    assertSameFile,
    getTestTempPath,
    logIterable,
    setupTestSuport,
    writeListToFile,
)

from hdl_checker import CACHE_NAME, WORK_PATH
from hdl_checker.builders.fallback import Fallback
from hdl_checker.diagnostics import (
    CheckerDiagnostic,
    DiagType,
    LibraryShouldBeOmited,
    ObjectIsNeverUsed,
    PathNotInProjectFile,
)
from hdl_checker.parsers.elements.dependency_spec import DependencySpec
from hdl_checker.parsers.elements.identifier import Identifier
from hdl_checker.path import Path
from hdl_checker.types import (
    BuildFlagScope,
    FileType,
    RebuildLibraryUnit,
    RebuildPath,
    RebuildUnit,
)
from hdl_checker.utils import ON_WINDOWS, removeIfExists

_logger = logging.getLogger(__name__)

TEST_TEMP_PATH = getTestTempPath(__name__)
TEST_PROJECT = p.join(TEST_TEMP_PATH, "test_project")

if six.PY2:
    FileNotFoundError = (  # pylint: disable=redefined-builtin,invalid-name
        IOError,
        OSError,
    )


class _SourceMock(SourceMock):
    base_path = TEST_TEMP_PATH


def _path(*args):
    # type: (str) -> str
    "Helper to reduce foorprint of p.join(TEST_TEMP_PATH, *args)"
    return p.join(TEST_TEMP_PATH, *args)


def _Path(*args):
    # type: (str) -> Path
    return Path(_path(*args))


def patchClassMap(**kwargs):
    import hdl_checker

    class_map = hdl_checker.serialization.CLASS_MAP.copy()
    for name, value in kwargs.items():
        class_map.update({name: value})

    return patch("hdl_checker.serialization.CLASS_MAP", class_map)


def _makeConfigFromDict(dict_):
    # type: (...) -> str
    filename = p.join(TEST_TEMP_PATH, "mock.prj")
    json.dump(dict_, open(filename, "w"))
    return filename


such.unittest.TestCase.maxDiff = None

with such.A("hdl_checker project") as it:
    if six.PY2:
        it.assertCountEqual = assertCountEqual(it)

    it.assertSameFile = assertSameFile(it)

    def _assertMsgQueueIsEmpty(project):
        msg = []
        while not project._msg_queue.empty():
            msg += [str(project._msg_queue.get())]

        if msg:
            msg.insert(
                0, "Message queue should be empty but has %d messages" % len(msg)
            )
            it.fail("\n".join(msg))

    it.assertMsgQueueIsEmpty = _assertMsgQueueIsEmpty

    with it.having("non existing root dir"):

        @it.has_setup
        def setup():
            it.project_file = Path("non_existing_file")
            it.assertFalse(p.exists(it.project_file.name))

        @it.should("raise exception when trying to instantiate")
        def test():
            project = StandaloneProjectBuilder(_Path("nonexisting"))
            with it.assertRaises(FileNotFoundError):
                project.setConfig(str(it.project_file))

    with it.having("no project file at all"):

        @it.has_setup
        def setup():
            it.project = StandaloneProjectBuilder(Path(TEST_PROJECT))

        @it.should("use fallback to Fallback builder")  # type: ignore
        def test():
            it.assertIsInstance(it.project.builder, Fallback)

        @it.should("still report static messages")  # type: ignore
        def test():

            _logger.info("Files: %s", it.project.database._paths)

            source = _SourceMock(
                filename=_path("some_lib_target.vhd"),
                library="some_lib",
                design_units=[{"name": "target", "type": "entity"}],
                dependencies={("work", "foo")},
            )

            it.assertCountEqual(
                it.project.getMessagesByPath(source.filename),
                {
                    LibraryShouldBeOmited(
                        library="work",
                        filename=_Path("some_lib_target.vhd"),
                        line_number=1,
                        column_number=9,
                    )
                },
            )

    with it.having("an existing and valid project file"):

        @it.has_setup
        def setup():
            setupTestSuport(TEST_TEMP_PATH)

            it.project = StandaloneProjectBuilder(_Path(TEST_TEMP_PATH))

            it.config_file = _makeConfigFromDict(
                {
                    "builder": MockBuilder.builder_name,
                    FileType.vhdl.value: {
                        "flags": {
                            BuildFlagScope.all.value: (
                                "-globalvhdl",
                                "-global-vhdl-flag",
                            ),
                            BuildFlagScope.dependencies.value: ("--vhdl-batch",),
                            BuildFlagScope.single.value: ("-single_build_flag_0",),
                        }
                    },
                    FileType.verilog.value: {
                        "flags": {
                            BuildFlagScope.all.value: (
                                "-globalverilog",
                                "-global-verilog-flag",
                            ),
                            BuildFlagScope.dependencies.value: ("--verilog-batch",),
                            BuildFlagScope.single.value: ("-single_build_flag_0",),
                        }
                    },
                    FileType.systemverilog.value: {
                        "flags": {
                            BuildFlagScope.all.value: (
                                "-globalsystemverilog",
                                "-global-systemverilog-flag",
                            ),
                            BuildFlagScope.dependencies.value: (
                                "--systemverilog-batch",
                            ),
                            BuildFlagScope.single.value: ("-single_build_flag_0",),
                        }
                    },
                }
            )

            it.project.setConfig(it.config_file)

        @it.should("use MockBuilder builder")  # type: ignore
        def test():
            with PatchBuilder():
                it.assertIsInstance(it.project.builder, MockBuilder)

        @it.should("save cache after checking a source")  # type: ignore
        def test():
            source = _SourceMock(
                library="some_lib", design_units=[{"name": "target", "type": "entity"}]
            )

            with patch("hdl_checker.base_server.json.dump", spec=json.dump) as func:
                it.project.getMessagesByPath(source.filename)
                func.assert_called_once()

        @it.should("recover from cache")  # type: ignore
        @patchClassMap(MockBuilder=MockBuilder)
        @patch("hdl_checker.base_server.HdlCodeCheckerBase._setState")
        def test(set_state):
            source = _SourceMock(
                library="some_lib", design_units=[{"name": "target", "type": "entity"}]
            )

            it.project.getMessagesByPath(source.filename)

            _ = StandaloneProjectBuilder(_Path(TEST_TEMP_PATH))

            set_state.assert_called_once()

            # Setting the config file should not trigger reparsing
            with patch.object(it.project, "_readConfig") as read_config:
                with patch(
                    "hdl_checker.base_server.WatchedFile.__init__", side_effect=[None]
                ) as watched_file:
                    old = it.project.config_file
                    it.project.setConfig(it.config_file)
                    it.project._updateConfigIfNeeded()
                    it.assertEqual(it.project.config_file, old)
                    read_config.assert_not_called()
                    watched_file.assert_not_called()

        @it.should("clean up root dir")  # type: ignore
        @patchClassMap(MockBuilder=MockBuilder)
        def test():
            if not ON_WINDOWS:
                it.assertMsgQueueIsEmpty(it.project)

            it.project.clean()

            project = StandaloneProjectBuilder(_Path(TEST_TEMP_PATH))

            if not ON_WINDOWS:
                it.assertMsgQueueIsEmpty(project)

            # Reset the project to the previous state
            setup()

        @it.should("warn when failing to recover from cache")  # type: ignore
        def test():
            it.project._saveCache()
            # Copy parameters of the object we're checking against
            root_dir = it.project.root_dir
            cache_filename = it.project._getCacheFilename()

            #  it.assertIsInstance(it.project.builder, Fallback)

            # Corrupt the cache file
            open(cache_filename.name, "w").write("corrupted cache contents")

            # Try to recreate
            project = StandaloneProjectBuilder(root_dir)

            if six.PY2:
                it.assertIn(
                    (
                        "warning",
                        "Unable to recover cache from '{}': "
                        "No JSON object could be decoded".format(cache_filename),
                    ),
                    list(project.getUiMessages()),
                )
            else:
                it.assertIn(
                    (
                        "warning",
                        "Unable to recover cache from '{}': "
                        "Expecting value: line 1 column 1 (char 0)".format(
                            cache_filename
                        ),
                    ),
                    list(project.getUiMessages()),
                )

            it.assertIsInstance(project.builder, Fallback)

        @it.should("get builder messages by path")  # type: ignore
        # Avoid saving to cache because the patched method is not JSON
        # serializable
        @patch("hdl_checker.base_server.json.dump")
        def test(_):
            with PatchBuilder():
                it.project.setConfig(Path(p.join(TEST_PROJECT, "vimhdl.prj")))
                it.project._readConfig()

            entity_a = _SourceMock(
                filename=_path("entity_a.vhd"),
                library="some_lib",
                design_units=[{"name": "entity_a", "type": "entity"}],
                dependencies={("work", "foo")},
            )

            path_to_foo = Path(p.join(TEST_PROJECT, "another_library", "foo.vhd"))

            diags = {
                # entity_a.vhd is the path we're compiling, inject a diagnostic
                # from the builder
                str(entity_a.filename): [
                    [
                        CheckerDiagnostic(
                            filename=entity_a.filename, checker=None, text="some text"
                        )
                    ]
                ],
                # foo.vhd is on the build sequence, check that diagnostics from
                # the build sequence are only included when their severity is
                # DiagType.ERROR
                str(path_to_foo): [
                    [
                        CheckerDiagnostic(
                            filename=path_to_foo,
                            checker=None,
                            text="should not be included",
                            severity=DiagType.WARNING,
                        ),
                        CheckerDiagnostic(
                            filename=path_to_foo,
                            checker=None,
                            text="style error should be included",
                            severity=DiagType.STYLE_ERROR,
                        ),
                        CheckerDiagnostic(
                            filename=path_to_foo,
                            checker=None,
                            text="should be included",
                            severity=DiagType.ERROR,
                        ),
                    ]
                ],
            }

            def build(  # pylint: disable=unused-argument
                path, library, scope, forced=False
            ):
                _logger.debug("Building library=%s, path=%s", library, path)
                path_diags = diags.get(str(path), [])
                if path_diags:
                    return path_diags.pop(), []
                return [], []

            with patch.object(it.project.builder, "build", build):
                _logger.info("Project paths: %s", it.project.database._paths)
                messages = list(it.project.getMessagesByPath(entity_a.filename))
                logIterable("Messages", messages, _logger.info)
                it.assertCountEqual(
                    messages,
                    [
                        LibraryShouldBeOmited(
                            library="work",
                            filename=entity_a.filename,
                            column_number=9,
                            line_number=1,
                        ),
                        PathNotInProjectFile(entity_a.filename),
                        CheckerDiagnostic(
                            filename=entity_a.filename, checker=None, text="some text"
                        ),
                        CheckerDiagnostic(
                            filename=path_to_foo,
                            checker=None,
                            text="style error should be included",
                            severity=DiagType.STYLE_ERROR,
                        ),
                        CheckerDiagnostic(
                            filename=path_to_foo,
                            checker=None,
                            text="should be included",
                            severity=DiagType.ERROR,
                        ),
                    ],
                )

        @it.should("Resolve dependency to path")  # type: ignore
        def test():
            path = _Path(TEST_PROJECT, "another_library", "foo.vhd")

            clock_divider = DependencySpec(
                name=Identifier("clock_divider"),
                library=Identifier("basic_library"),
                owner=path,
                locations=(),
            )

            it.assertEqual(
                it.project.resolveDependencyToPath(clock_divider),
                (
                    _Path(TEST_PROJECT, "basic_library", "clock_divider.vhd"),
                    Identifier("basic_library"),
                ),
            )

        @it.should("Not resolve dependencies whose library is built in")  # type: ignore
        def test():
            path = _Path(TEST_PROJECT, "another_library", "foo.vhd")

            numeric_std = DependencySpec(
                name=Identifier("numeric_std"),
                library=Identifier("ieee"),
                owner=path,
                locations=(),
            )

            it.assertIs(it.project.resolveDependencyToPath(numeric_std), None)

        @it.should(  # type: ignore
            "warn when unable to recreate a builder described in cache"
        )
        @patch(
            "hdl_checker.base_server.getBuilderByName", new=lambda name: FailingBuilder
        )
        def test():
            if ON_WINDOWS:
                raise it.skipTest("Test doesn't run on Windows")

            cache_content = {"builder": FailingBuilder.builder_name}

            cache_path = it.project._getCacheFilename()
            if p.exists(p.dirname(cache_path.name)):
                shutil.rmtree(p.dirname(cache_path.name))

            os.makedirs(p.dirname(cache_path.name))

            with open(cache_path.name, "w") as fd:
                fd.write(repr(cache_content))

            found = True
            while not it.project._msg_queue.empty():
                severity, message = it.project._msg_queue.get()
                _logger.info("Message found: [%s] %s", severity, message)
                if (
                    message
                    == "Failed to create builder '%s'" % FailingBuilder.builder_name
                ):
                    found = True
                    break

            it.assertTrue(found, "Failed to warn that cache recovering has failed")
            it.assertTrue(it.project.builder.builder_name, "Fallback")

    with it.having("test_project as reference and a valid project file"):

        @it.has_setup
        def setup():
            setupTestSuport(TEST_TEMP_PATH)

            removeIfExists(p.join(TEST_TEMP_PATH, WORK_PATH, CACHE_NAME))

            with PatchBuilder():
                it.project = StandaloneProjectBuilder(_Path(TEST_TEMP_PATH))
                it.project.setConfig(Path(p.join(TEST_PROJECT, "vimhdl.prj")))
                it.project._updateConfigIfNeeded()

            from pprint import pformat

            _logger.info(
                "Database state:\n%s", pformat(it.project.database.__jsonEncode__())
            )

            it.assertTrue(it.project.database.paths)
            it.assertIsInstance(it.project.builder, MockBuilder)

        @it.should("use mock builder")  # type: ignore
        def test():
            it.assertIsInstance(it.project.builder, MockBuilder)

        @it.should("get messages for an absolute path")  # type: ignore
        def test():
            filename = p.join(TEST_PROJECT, "another_library", "foo.vhd")

            if not ON_WINDOWS:
                it.assertMsgQueueIsEmpty(it.project)

            diagnostics = it.project.getMessagesByPath(Path(filename))

            it.assertIn(
                ObjectIsNeverUsed(
                    filename=Path(filename),
                    line_number=29,
                    column_number=12,
                    object_type="signal",
                    object_name="neat_signal",
                ),
                diagnostics,
            )

            if not ON_WINDOWS:
                it.assertMsgQueueIsEmpty(it.project)

            it.assertTrue(it.project.database.paths)

        @it.should("get messages for relative path")  # type: ignore
        def test():
            filename = p.relpath(
                p.join(TEST_PROJECT, "another_library", "foo.vhd"),
                str(it.project.root_dir),
            )

            it.assertFalse(p.isabs(filename))

            if not ON_WINDOWS:
                it.assertMsgQueueIsEmpty(it.project)

            diagnostics = it.project.getMessagesByPath(Path(filename))

            it.assertIn(
                ObjectIsNeverUsed(
                    filename=Path(p.join(TEST_PROJECT, "another_library", "foo.vhd")),
                    line_number=29,
                    column_number=12,
                    object_type="signal",
                    object_name="neat_signal",
                ),
                diagnostics,
            )

            if not ON_WINDOWS:
                it.assertMsgQueueIsEmpty(it.project)

        @it.should("get messages with text")  # type: ignore
        def test():
            it.assertTrue(it.project.database.paths)

            filename = Path(p.join(TEST_PROJECT, "another_library", "foo.vhd"))
            original_content = open(filename.name, "r").read().split("\n")

            content = "\n".join(
                original_content[:28]
                + ["signal another_signal : std_logic;"]
                + original_content[28:]
            )

            _logger.debug("File content")
            for lnum, line in enumerate(content.split("\n")):
                _logger.debug("%2d| %s", (lnum + 1), line)

            if not ON_WINDOWS:
                it.assertMsgQueueIsEmpty(it.project)

            diagnostics = set(it.project.getMessagesWithText(filename, content))

            logIterable("Diagnostics", diagnostics, _logger.info)

            it.assertTrue(it.project.config_file)

            expected = [
                ObjectIsNeverUsed(
                    filename=filename,
                    line_number=30,
                    column_number=12,
                    object_type="signal",
                    object_name="neat_signal",
                ),
                ObjectIsNeverUsed(
                    filename=filename,
                    line_number=29,
                    column_number=8,
                    object_type="signal",
                    object_name="another_signal",
                ),
            ]

            it.assertCountEqual(diagnostics, expected)
            if not ON_WINDOWS:
                it.assertMsgQueueIsEmpty(it.project)

        @it.should(  # type: ignore
            "get messages with text for file outside the project file"
        )
        def test():
            filename = Path(p.join(TEST_TEMP_PATH, "some_file.vhd"))
            writeListToFile(str(filename), ["entity some_entity is end;"])

            content = "\n".join(
                ["library work;", "use work.all;", "entity some_entity is end;"]
            )

            if not ON_WINDOWS:
                it.assertMsgQueueIsEmpty(it.project)

            diagnostics = it.project.getMessagesWithText(filename, content)

            _logger.debug("Records received:")
            for diagnostic in diagnostics:
                _logger.debug("- %s", diagnostic)

            expected = [
                LibraryShouldBeOmited(
                    library="work", filename=filename, column_number=9, line_number=1
                ),
                PathNotInProjectFile(filename),
            ]

            try:
                it.assertCountEqual(expected, diagnostics)
            except:
                _logger.warning("Expected:")
                for exp in expected:
                    _logger.warning(exp)

                raise

            if not ON_WINDOWS:
                it.assertMsgQueueIsEmpty(it.project)

        @it.should("get updated messages")  # type: ignore
        def test():
            filename = Path(p.join(TEST_PROJECT, "another_library", "foo.vhd"))

            if not ON_WINDOWS:
                it.assertMsgQueueIsEmpty(it.project)

            code = open(str(filename), "r").read().split("\n")

            code[28] = "-- " + code[28]

            writeListToFile(str(filename), code)

            diagnostics = it.project.getMessagesByPath(filename)

            try:
                it.assertNotIn(
                    ObjectIsNeverUsed(
                        object_type="constant",
                        object_name="ADDR_WIDTH",
                        line_number=29,
                        column_number=14,
                    ),
                    diagnostics,
                )
            finally:
                # Remove the comment we added
                code[28] = code[28][3:]
                writeListToFile(str(filename), code)

            if not ON_WINDOWS:
                it.assertMsgQueueIsEmpty(it.project)

        @it.should("get messages by path of a different source")  # type: ignore
        def test():
            filename = Path(p.join(TEST_PROJECT, "basic_library", "clock_divider.vhd"))

            if not ON_WINDOWS:
                it.assertMsgQueueIsEmpty(it.project)

            it.assertCountEqual(
                it.project.getMessagesByPath(filename),
                [
                    ObjectIsNeverUsed(
                        filename=filename,
                        line_number=27,
                        column_number=12,
                        object_type="signal",
                        object_name="clk_enable_unused",
                    )
                ],
            )

            if not ON_WINDOWS:
                it.assertMsgQueueIsEmpty(it.project)

        @it.should(  # type: ignore
            "get messages from a source outside the project file"
        )
        def test():
            filename = Path(p.join(TEST_TEMP_PATH, "some_file.vhd"))
            writeListToFile(str(filename), ["library some_lib;"])

            if not ON_WINDOWS:
                it.assertMsgQueueIsEmpty(it.project)

            diagnostics = it.project.getMessagesByPath(filename)

            _logger.info("Records found:")
            for diagnostic in diagnostics:
                _logger.info(diagnostic)

            it.assertIn(PathNotInProjectFile(filename), diagnostics)

            # The builder should find other issues as well...
            it.assertTrue(
                len(diagnostics) > 1,
                "It was expected that the builder added some "
                "message here indicating an error",
            )

            if not ON_WINDOWS:
                it.assertMsgQueueIsEmpty(it.project)

        def basicRebuildTest(test_filename, rebuilds):
            calls = []
            ret_list = list(reversed(rebuilds))

            # Rebuild formats are:
            # - {unit_type: '', 'unit_name': }
            # - {library_name: '', 'unit_name': }
            # - {rebuild_path: ''}
            def _buildAndParse(_, path, library, forced=False):
                _logger.warning("%s, %s, %s", path, library, forced)
                calls.append(str(path))
                if ret_list:
                    return [], ret_list.pop()
                return [], []

            with patch.object(MockBuilder, "_buildAndParse", _buildAndParse):
                it.assertFalse(
                    list(it.project._getBuilderMessages(Path(test_filename)))
                )

            it.assertFalse(ret_list, "Some rebuilds were not used: {}".format(ret_list))

            return calls

        def _RebuildLibraryUnit(name, library):
            return RebuildLibraryUnit(
                name=Identifier(name), library=Identifier(library)
            )

        @it.should(  # type: ignore
            "rebuild sources when needed within the same library"
        )
        def test():
            it.project.database._clearLruCaches()
            filename = p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd")
            rebuilds = [
                [_RebuildLibraryUnit(name="clock_divider", library="basic_library")]
            ]

            logIterable("Design units", it.project.database.design_units, _logger.info)
            calls = basicRebuildTest(filename, rebuilds)

            # Calls should be
            # - first to build the source we wanted
            # - second to build the file we said needed to be rebuilt
            # - third should build the original source after handling a rebuild
            it.assertEqual(
                calls,
                [
                    p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd"),
                    p.join(TEST_PROJECT, "basic_library", "clock_divider.vhd"),
                    p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd"),
                ],
            )

        @it.should(  # type: ignore
            "rebuild sources when changing a package on different libraries"
        )
        def test():
            filename = p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd")
            rebuilds = [[_RebuildLibraryUnit(library="another_library", name="foo")]]

            calls = basicRebuildTest(filename, rebuilds)

            # Calls should be
            # - first to build the source we wanted
            # - second to build the file we said needed to be rebuilt
            # - third should build the original source after handling a rebuild
            it.assertEqual(
                calls,
                [
                    p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd"),
                    p.join(TEST_PROJECT, "another_library", "foo.vhd"),
                    p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd"),
                ],
            )

        def _RebuildPath(path):
            return RebuildPath(Path(path))

        @it.should("rebuild sources with path as a hint")
        def test():
            filename = p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd")

            abs_path = p.join(
                TEST_PROJECT, "basic_library", "package_with_constants.vhd"
            )

            # Force relative path as well
            rel_path = p.relpath(abs_path, str(it.project.root_dir))

            it.assertFalse(p.isabs(rel_path))

            rebuilds = [[_RebuildPath(rel_path)]]

            calls = basicRebuildTest(filename, rebuilds)

            #  Calls should be
            #  - first to build the source we wanted
            #  - second to build the file we said needed to be rebuilt
            #  - third should build the original source after handling a rebuild
            it.assertEqual(
                calls,
                [
                    p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd"),
                    abs_path,
                    p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd"),
                ],
            )

        def _RebuildUnit(name, type_):
            return RebuildUnit(name=Identifier(name), type_=Identifier(type_))

        @it.should("rebuild package if needed")  # type: ignore
        def test():
            filename = p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd")

            # - {unit_type: '', 'unit_name': }
            rebuilds = [[_RebuildUnit(name="very_common_pkg", type_="package")]]

            calls = basicRebuildTest(filename, rebuilds)

            # Calls should be
            # - first to build the source we wanted
            # - second to build the file we said needed to be rebuilt
            # - third should build the original source after handling a rebuild
            it.assertEqual(
                calls,
                [
                    p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd"),
                    p.join(TEST_PROJECT, "basic_library", "very_common_pkg.vhd"),
                    p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd"),
                ],
            )

        @it.should("rebuild a combination of all")  # type: ignore
        def test():
            filename = p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd")

            # - {unit_type: '', 'unit_name': }
            rebuilds = [
                [
                    _RebuildUnit(name="very_common_pkg", type_="package"),
                    _RebuildPath(
                        p.join(
                            TEST_PROJECT, "basic_library", "package_with_constants.vhd"
                        )
                    ),
                    _RebuildLibraryUnit(name="foo", library="another_library"),
                ]
            ]

            calls = basicRebuildTest(filename, rebuilds)

            # Calls should be
            # - first to build the source we wanted
            # - second to build the file we said needed to be rebuilt
            # - third should build the original source after handling a rebuild
            it.assertEqual(
                calls,
                [
                    p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd"),
                    p.join(TEST_PROJECT, "basic_library", "very_common_pkg.vhd"),
                    p.join(TEST_PROJECT, "basic_library", "package_with_constants.vhd"),
                    p.join(TEST_PROJECT, "another_library", "foo.vhd"),
                    p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd"),
                ],
            )

        @it.should("give up trying to rebuild after 20 attempts")  # type: ignore
        def test():
            filename = p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd")

            # - {unit_type: '', 'unit_name': }
            rebuilds = 20 * [
                [_RebuildLibraryUnit(name="foo", library="another_library")],
                [],
            ]

            calls = basicRebuildTest(filename, rebuilds)

            # Calls should be
            # - first to build the source we wanted
            # - second to build the file we said needed to be rebuilt
            # - third should build the original source after handling a rebuild
            it.assertEqual(
                calls,
                20
                * [
                    p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd"),
                    p.join(TEST_PROJECT, "another_library", "foo.vhd"),
                ],
            )

            # Not sure why this is needed, might be hiding something weird
            time.sleep(0.1)

            ui_msgs = list(it.project.getUiMessages())
            logIterable("UI messages", ui_msgs, _logger.info)

            it.assertIn(
                ("error", "Unable to build '{}' after 20 attempts".format(filename)),
                ui_msgs,
            )


it.createTests(globals())



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=function-redefined
# pylint: disable=missing-docstring
# pylint: disable=protected-access

import logging
import os
import os.path as p
import shutil
from typing import Any, List

import mock
import parameterized  # type: ignore
import unittest2  # type: ignore

from hdl_checker.tests import (
    SourceMock,
    TestCase,
    getTestTempPath,
    parametrizeClassWithBuilders,
    setupTestSuport,
)

from hdl_checker.builder_utils import (
    AVAILABLE_BUILDERS,
    GHDL,
    XVHDL,
    AnyBuilder,
    Fallback,
    MSim,
)
from hdl_checker.diagnostics import BuilderDiag, DiagType
from hdl_checker.exceptions import SanityCheckError
from hdl_checker.parsers.elements.dependency_spec import DependencySpec
from hdl_checker.parsers.elements.identifier import Identifier
from hdl_checker.path import Path
from hdl_checker.types import (
    BuildFlagScope,
    DesignUnitType,
    FileType,
    RebuildLibraryUnit,
    RebuildPath,
    RebuildUnit,
)

_logger = logging.getLogger(__name__)

TEST_TEMP_PATH = getTestTempPath(__name__)
SOURCES_PATH = p.join(TEST_TEMP_PATH, "test_builders")

BUILDER_CLASS_MAP = {"msim": MSim, "xvhdl": XVHDL, "ghdl": GHDL, "fallback": Fallback}


class _SourceMock(SourceMock):
    base_path = TEST_TEMP_PATH


def _source(*args):
    # type: (str) -> Path
    "Helper to reduce foorprint of Path(p.join(SOURCES_PATH, *args))"
    return Path(p.join(SOURCES_PATH, *args))


def _temp(*args):
    # type: (str) -> Path
    "Helper to reduce foorprint of Path(p.join(TEST_TEMP_PATH, *args))"
    return Path(p.join(TEST_TEMP_PATH, *args))


@parametrizeClassWithBuilders
class TestBuilder(TestCase):
    # Create defaults so that pylint doesn't complain about non existing
    # members
    builder_name = None
    builder_path = None

    @classmethod
    def setUpClass(cls):
        setupTestSuport(TEST_TEMP_PATH)

    def setUp(self):
        # type: (...) -> Any
        # Add builder path to the env
        self.original_env = os.environ.copy()

        # Add the builder path to the environment so we can call it
        if self.builder_path:
            _logger.info("Adding '%s' to the system path", self.builder_path)
            self.patch = mock.patch.dict(
                "os.environ",
                {"PATH": os.pathsep.join([self.builder_path, os.environ["PATH"]])},
            )
            self.patch.start()

        assert self.builder_name is not None  # To make mypy happy

        builder_class = BUILDER_CLASS_MAP[self.builder_name]
        work_folder = _temp("_%s" % self.builder_name)
        _logger.info("Builder class: %s, work folder is %s", builder_class, work_folder)
        self.builder = builder_class(work_folder, mock.MagicMock())  # type: AnyBuilder
        self.builder_class = builder_class

    def tearDown(self):
        # type: (...) -> Any
        if self.builder_path:
            self.patch.stop()
        if p.exists("._%s" % self.builder_name):
            shutil.rmtree("._%s" % self.builder_name)

    def test_environment_check(self):
        # type: (...) -> Any
        self.builder.checkEnvironment()

    def test_builder_reports_its_available(self):  # pylint: disable=invalid-name
        # type: (...) -> Any
        self.assertTrue(self.builder_class.isAvailable())  # type: ignore

    def test_create_library_multiple_times(self):  # pylint: disable=invalid-name
        # type: (...) -> Any
        self.builder._createLibraryIfNeeded(Identifier("random_lib"))
        self.builder._createLibraryIfNeeded(Identifier("random_lib"))

    def test_builder_does_nothing_when_creating_builtin_libraries(
        self
    ):  # pylint: disable=invalid-name
        # type: (...) -> Any
        #  pre = open(self.builder._xvhdlini).read()
        self.builder._createLibraryIfNeeded(Identifier("ieee"))
        #  post = open(self.builder._xvhdlini).read()

    def test_finds_builtin_libraries(self):
        # type: (...) -> Any
        expected = []  # type: List[str]

        if not isinstance(self.builder, Fallback):
            expected += ["ieee", "std"]

        if isinstance(self.builder, MSim):
            expected += ["modelsim_lib"]

        for lib in map(Identifier, expected):
            self.assertIn(lib, self.builder.builtin_libraries)

    @parameterized.parameterized.expand(
        [
            ("/some/file/with/abs/path.vhd",),
            ("some/file/with/relative/path.vhd",),
            ("some_file_on_same_level.vhd",),
            (r"C:\some\file\on\windows.vhd",),
        ]
    )
    def test_parse_msim_result(self, path):
        # type: (...) -> Any
        if not isinstance(self.builder, MSim):
            raise unittest2.SkipTest("ModelSim only test")

        self.assertEqual(
            list(
                self.builder._makeRecords(
                    '** Error: %s(21): near "EOF": (vcom-1576) ' "expecting ';'." % path
                )
            ),
            [
                BuilderDiag(
                    builder_name=self.builder_name,
                    text="near \"EOF\": expecting ';'.",
                    filename=Path(path),
                    line_number=21,
                    error_code="vcom-1576",
                    severity=DiagType.ERROR,
                )
            ],
        )

        self.assertEqual(
            list(
                self.builder._makeRecords(
                    "** Warning: %s(23): (vcom-1320) Type of expression "
                    "\"(OTHERS => '0')\" is ambiguous; using element type "
                    "STD_LOGIC_VECTOR, not aggregate type register_type." % path
                )
            ),
            [
                BuilderDiag(
                    builder_name=self.builder_name,
                    text="Type of expression \"(OTHERS => '0')\" is "
                    "ambiguous; using element type STD_LOGIC_VECTOR, not "
                    "aggregate type register_type.",
                    filename=Path(path),
                    line_number=23,
                    error_code="vcom-1320",
                    severity=DiagType.WARNING,
                )
            ],
        )

        self.assertEqual(
            list(
                self.builder._makeRecords(
                    "** Warning: %s(39): (vcom-1514) Range choice direction "
                    "(downto) does not determine aggregate index range "
                    "direction (to)." % path
                )
            ),
            [
                BuilderDiag(
                    builder_name=self.builder_name,
                    text="Range choice direction (downto) does not determine "
                    "aggregate index range direction (to).",
                    filename=Path(path),
                    line_number=39,
                    error_code="vcom-1514",
                    severity=DiagType.WARNING,
                )
            ],
        )

        self.assertEqual(
            list(
                self.builder._makeRecords(
                    "** Error: (vcom-11) Could not find work.regfile_pkg."
                )
            ),
            [
                BuilderDiag(
                    builder_name=self.builder_name,
                    text="Could not find work.regfile_pkg.",
                    error_code="vcom-11",
                    severity=DiagType.ERROR,
                )
            ],
        )

        self.assertEqual(
            list(
                self.builder._makeRecords(
                    "** Error (suppressible): %s(7): (vcom-1195) Cannot find "
                    'expanded name "work.regfile_pkg".' % path
                )
            ),
            [
                BuilderDiag(
                    builder_name=self.builder_name,
                    text='Cannot find expanded name "work.regfile_pkg".',
                    filename=Path(path),
                    line_number=7,
                    error_code="vcom-1195",
                    severity=DiagType.ERROR,
                )
            ],
        )

        self.assertEqual(
            list(
                self.builder._makeRecords(
                    "** Error: %s(7): Unknown expanded name." % path
                )
            ),
            [
                BuilderDiag(
                    builder_name=self.builder_name,
                    text="Unknown expanded name.",
                    line_number="7",
                    filename=Path(path),
                    severity=DiagType.ERROR,
                )
            ],
        )

        self.assertEqual(
            list(
                self.builder._makeRecords(
                    "** Warning: [14] %s(103): (vcom-1272) Length of expected "
                    "is 4; length of actual is 8." % path
                )
            ),
            [
                BuilderDiag(
                    builder_name=self.builder_name,
                    text="Length of expected is 4; length of actual is 8.",
                    line_number="103",
                    error_code="vcom-1272",
                    filename=Path(path),
                    severity=DiagType.WARNING,
                )
            ],
        )

        self.assertEqual(
            list(
                self.builder._makeRecords(
                    "** Warning: [14] %s(31): (vcom-1246) Range -1 downto 0 "
                    "is null." % path
                )
            ),
            [
                BuilderDiag(
                    builder_name=self.builder_name,
                    text="Range -1 downto 0 is null.",
                    line_number="31",
                    error_code="vcom-1246",
                    filename=Path(path),
                    severity=DiagType.WARNING,
                )
            ],
        )

    @parameterized.parameterized.expand(
        [
            ("/some/file/with/abs/path.vhd",),
            ("some/file/with/relative/path.vhd",),
            ("some_file_on_same_level.vhd",),
            (r"C:\some\file\on\windows.vhd",),
        ]
    )
    def test_parse_ghdl_result(self, path):
        # type: (...) -> Any
        if not isinstance(self.builder, GHDL):
            raise unittest2.SkipTest("GHDL only test")

        records = list(
            self.builder._makeRecords(
                "%s:11:35: extra ';' at end of interface list" % path
            )
        )

        expected = [
            BuilderDiag(
                builder_name=self.builder_name,
                filename=Path(path),
                line_number=11,
                column_number=35,
                severity=DiagType.ERROR,
                text="extra ';' at end of interface list",
            )
        ]

        self.assertCountEqual(records, expected)

    @parameterized.parameterized.expand(
        [
            ("/some/file/with/abs/path.vhd",),
            ("some/file/with/relative/path.vhd",),
            ("some_file_on_same_level.vhd",),
        ]
    )
    def test_parse_xvhdl_result(self, path):
        # type: (...) -> Any
        if not isinstance(self.builder, XVHDL):
            raise unittest2.SkipTest("XVHDL only test")

        self.assertEqual(
            list(
                self.builder._makeRecords(
                    "ERROR: [VRFC 10-1412] syntax error near ) [%s:12]" % path
                )
            ),
            [
                BuilderDiag(
                    builder_name=self.builder_name,
                    text="syntax error near )",
                    filename=Path(path),
                    line_number=12,
                    error_code="VRFC 10-1412",
                    severity=DiagType.ERROR,
                )
            ],
        )

    def test_vhdl_compilation(self):
        # type: (...) -> Any
        if FileType.vhdl not in self.builder.file_types:
            raise unittest2.SkipTest(
                "Builder {} doesn't support VHDL".format(self.builder_name)
            )

        source = _source("no_messages.vhd")
        with mock.patch.object(
            self.builder._database, "getLibrary", return_value=Identifier("work")
        ):
            records, rebuilds = self.builder.build(
                source, Identifier("work"), BuildFlagScope.single
            )
        self.assertNotIn(
            DiagType.ERROR,
            [x.severity for x in records],
            "This source should not generate errors.",
        )
        self.assertFalse(rebuilds)

    def test_verilog_compilation(self):
        # type: (...) -> Any
        if FileType.verilog not in self.builder.file_types:
            raise unittest2.SkipTest(
                "Builder {} doesn't support Verilog".format(self.builder_name)
            )

        source = _source("no_messages.v")
        with mock.patch.object(
            self.builder._database, "getLibrary", return_value=Identifier("work")
        ):
            records, rebuilds = self.builder.build(
                source, Identifier("work"), BuildFlagScope.single
            )
        self.assertNotIn(
            DiagType.ERROR,
            [x.severity for x in records],
            "This source should not generate errors.",
        )
        self.assertFalse(rebuilds)

    def test_systemverilog_compilation(self):
        # type: (...) -> Any
        if FileType.systemverilog not in self.builder.file_types:
            raise unittest2.SkipTest(
                "Builder {} doesn't support SystemVerilog".format(self.builder_name)
            )

        source = _source("no_messages.sv")

        with mock.patch.object(
            self.builder._database, "getLibrary", return_value=Identifier("work")
        ):
            records, rebuilds = self.builder.build(
                source, Identifier("work"), BuildFlagScope.single
            )

        self.assertNotIn(
            DiagType.ERROR,
            [x.severity for x in records],
            "This source should not generate errors.",
        )
        self.assertFalse(rebuilds)

    def test_catch_a_known_error(self):
        # type: (...) -> Any
        source = _source("source_with_error.vhd")

        records, rebuilds = self.builder.build(
            source, Identifier("lib"), forced=True, scope=BuildFlagScope.single
        )

        for record in records:
            _logger.info(record)

        if self.builder_name == "msim":
            expected = [
                BuilderDiag(
                    filename=source,
                    builder_name=self.builder_name,
                    text='Unknown identifier "some_lib".',
                    line_number=4,
                    error_code="vcom-1136",
                    severity=DiagType.ERROR,
                )
            ]
        elif self.builder_name == "ghdl":
            expected = [
                BuilderDiag(
                    filename=source,
                    builder_name=self.builder_name,
                    text='no declaration for "some_lib"',
                    line_number=4,
                    column_number=5,
                    severity=DiagType.ERROR,
                )
            ]
        elif self.builder_name == "xvhdl":
            # XVHDL reports different errors depending on the version
            expected = [
                BuilderDiag(
                    filename=source,
                    builder_name=self.builder_name,
                    text="some_lib is not declared",
                    line_number=4,
                    error_code="VRFC 10-91",
                    severity=DiagType.ERROR,
                ),
                BuilderDiag(
                    filename=source,
                    builder_name=self.builder_name,
                    text="'some_lib' is not declared",
                    line_number="4",
                    error_code="VRFC 10-2989",
                    severity=DiagType.ERROR,
                ),
            ]

        if not isinstance(self.builder, Fallback):
            self.assertEqual(
                len(records),
                1,
                "Expected only one record but got {}".format(len(records)),
            )
            self.assertIn(records.pop(), expected)
        else:
            self.assertFalse(records)

        self.assertFalse(rebuilds)

    def test_msim_recompile_msg_0(self):
        # type: (...) -> Any
        if not isinstance(self.builder, MSim):
            raise unittest2.SkipTest("ModelSim only test")

        line = (
            "** Error: (vcom-13) Recompile foo_lib.bar_component because "
            "foo_lib.foo_lib_pkg has changed."
        )

        self.assertEqual(
            [{"library_name": "foo_lib", "unit_name": "bar_component"}],
            self.builder._searchForRebuilds(line),
        )

    def test_msim_recompile_msg_1(self):
        # type: (...) -> Any
        if not isinstance(self.builder, MSim):
            raise unittest2.SkipTest("ModelSim only test")

        line = (
            "** Error: (vcom-13) Recompile foo_lib.bar_component because "
            "foo_lib.foo_lib_pkg, base_library.very_common_package have changed."
        )

        self.assertEqual(
            [{"library_name": "foo_lib", "unit_name": "bar_component"}],
            self.builder._searchForRebuilds(line),
        )

    def test_ghdl_recompile_msg(self):
        # type: (...) -> Any
        if not isinstance(self.builder, GHDL):
            raise unittest2.SkipTest("GHDL only test")

        line = 'somefile.vhd:12:13: package "leon3" is obsoleted by package "amba"'

        self.assertEqual(
            [{"unit_type": "package", "unit_name": "leon3"}],
            self.builder._searchForRebuilds(line),
        )

    # Rebuild formats are:
    # - {unit_type: '', 'unit_name': }
    # - {library_name: '', 'unit_name': }
    # - {rebuild_path: ''}
    @parameterized.parameterized.expand(
        [
            (
                {"unit_type": "package", "unit_name": "very_common_pkg"},
                RebuildUnit(
                    name=Identifier("very_common_pkg"), type_=DesignUnitType.package
                ),
            ),
            # Should replace 'work' with the path's library
            (
                {"library_name": "work", "unit_name": "foo"},
                RebuildLibraryUnit(
                    name=Identifier("foo"), library=Identifier("some_lib")
                ),
            ),
            # Should not touch the library name when != 'work'
            (
                {"library_name": "foo", "unit_name": "bar"},
                RebuildLibraryUnit(name=Identifier("bar"), library=Identifier("foo")),
            ),
            ({"rebuild_path": "some_path"}, RebuildPath(Path("some_path"))),
        ]
    )
    def test_get_rebuilds(self, rebuild_info, expected):
        # type: (...) -> Any
        _logger.info("Rebuild info is %s", rebuild_info)
        library = Identifier("some_lib", False)
        with mock.patch.object(
            self.builder, "_searchForRebuilds", return_value=[rebuild_info]
        ):
            self.builder._database.getDependenciesByPath = mock.MagicMock(
                return_value=[
                    DependencySpec(
                        owner=Path(""),
                        name=Identifier("very_common_pkg"),
                        library=Identifier("work"),
                    )
                ]
            )

            self.assertCountEqual(
                self.builder._getRebuilds(_source("source.vhd"), "", library),
                {expected},
            )


class TestSanityError(unittest2.TestCase):
    @parameterized.parameterized.expand([(x,) for x in AVAILABLE_BUILDERS])
    def test_not_available(self, builder_class):
        # type: (...) -> Any
        if builder_class is Fallback:
            self.assertTrue(builder_class.isAvailable())
        else:
            self.assertFalse(builder_class.isAvailable())

    @parameterized.parameterized.expand([(x,) for x in AVAILABLE_BUILDERS])
    def test_raises_sanity_error(self, builder_class):
        # type: (...) -> Any
        if builder_class is Fallback:
            raise self.skipTest("Fallback won't raise any exception")

        _logger.info("Testing builder %s", builder_class.builder_name)

        with self.assertRaises(SanityCheckError):
            _ = builder_class(
                Path(p.join(TEST_TEMP_PATH, "_%s" % builder_class.builder_name)), None
            )



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Common type definitions for type hinting"
from collections import namedtuple
from enum import Enum
from typing import NamedTuple, Optional, Tuple, Union

from hdl_checker.exceptions import UnknownTypeExtension
from hdl_checker.parsers.elements.identifier import Identifier
from hdl_checker.path import Path


class DesignUnitType(str, Enum):
    "Specifies tracked design unit types"
    package = "package"
    entity = "entity"
    context = "context"


BuildFlags = Tuple[str, ...]
LibraryAndUnit = namedtuple("LibraryAndUnit", ["library", "unit"])

RebuildUnit = NamedTuple(
    "RebuildUnit", (("name", Identifier), ("type_", DesignUnitType))
)
RebuildLibraryUnit = NamedTuple(
    "RebuildLibraryUnit", (("name", Identifier), ("library", Identifier))
)
RebuildPath = NamedTuple("RebuildPath", (("path", Path),))

RebuildInfo = Union[RebuildUnit, RebuildLibraryUnit, RebuildPath]


class FileType(Enum):
    "RTL file types"
    vhdl = "vhdl"
    verilog = "verilog"
    systemverilog = "systemverilog"

    @staticmethod
    def fromPath(path):
        # type: (Path) -> FileType
        "Extracts FileType from the given path's extension"
        ext = path.name.split(".")[-1].lower()
        if ext in ("vhd", "vhdl"):
            return FileType.vhdl
        if ext in ("v", "vh"):
            return FileType.verilog
        if ext in ("sv", "svh"):
            return FileType.systemverilog
        raise UnknownTypeExtension(path)

    def __jsonEncode__(self):
        """
        Gets a dict that describes the current state of this object
        """
        return {"value": self.name}

    @classmethod
    def __jsonDecode__(cls, state):
        """Returns an object of cls based on a given state"""
        return cls(state["value"])


class BuildFlagScope(Enum):
    """
    Scopes of a given set of flags. Values of the items control the actual
    fields extracted from the JSON config
    """

    single = "single"
    dependencies = "dependencies"
    all = "global"


class MarkupKind(Enum):
    "LSP Markup kinds"
    PlainText = "plaintext"
    Markdown = "markdown"


# A location on a source file
Location = NamedTuple("Location", (("line", Optional[int]), ("column", Optional[int])))

# A location range within a source file
Range = NamedTuple("Range", (("start", Location), ("end", Optional[Location])))
